# -*- coding: utf-8 -*-
#
# Copyright (c) 2009 by xt <xt@bash.no>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# (this script requires WeeChat 0.3.0 or newer)
#
# History:
# 2022-10-19, Guillermo Castro <github@codegeek.dev>
#   version 0.9: fix regex parsing of INVITE to allow non-ircv3 matching
# 2022-03-2, h-0-s-h <noreply@ic3.gov>
#   version 0.8: fix reged parsing of INVITE message to account for ircv3/CAPS style/etc
#                (@time=2022-03-02T19:00:30.041Z :XXXXX!~XXXXX@xxxxx INVITE h-0-s-h :#xxxxx)
# 2018-10-03, Pol Van Aubel <dev@polvanaubel.com>
#   version 0.7: Python3 compatibility. Considerations:
#                - Settings during registration are iterated over only once, so
#                  code should be succinct rather than efficient on both Py2/3.
#                - The autojoin_keys zip is turned into a dict, so wouldn't
#                  make sense to first turn into a list as futurize suggests.
# 2015-10-11, Simmo Saan <simmo.saan@gmail.com>
#   version 0.6: allow joining channels with keys in autojoin
# 2013-12-21, Sebastien Helleu <flashcode@flashtux.org>
#   version 0.5: fix parsing of INVITE message
# 2013-11-28, sakkemo <scajanus@gmail.com>
#   version 0.4: add whitelist for nicks/channels
# 2009-11-09, xt <xt@bash.no>
#   version 0.3: add ignore option for channels
# 2009-10-29, xt <xt@bash.no>
#   version 0.2: add ignore option
# 2009-10-28, xt <xt@bash.no>
#   version 0.1: initial release

import weechat as w
import re

SCRIPT_NAME    = "autojoin_on_invite"
SCRIPT_AUTHOR  = "xt <xt@bash.no>"
SCRIPT_VERSION = "0.9"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Auto joins channels when invited"

# script options
settings = {
        'whitelist_nicks': '', # comma separated list of nicks,
                               # overrides ignore_nicks
        'whitelist_channels': '', # comma separated list of channels,
                               # overrides ignore_channels
        'ignore_nicks': '', # comma separated list of nicks
                            #that we will not accept auto invite from
        'ignore_channels': '', # comma separated list of channels to not join
        'autojoin_key': 'on', # use channel keys from server's autojoin list
}

if w.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE,
                    SCRIPT_DESC, "", ""):
    for option, default_value in settings.items():
        if not w.config_is_set_plugin(option):
            w.config_set_plugin(option, default_value)

    w.hook_signal('*,irc_in2_invite', 'invite_cb', '')

def join(server, channel):
    key = None

    if w.config_string_to_boolean(w.config_get_plugin('autojoin_key')):
        autojoin = w.config_string(w.config_get('irc.server.%s.autojoin' % server)).split(' ', 1)

        if len(autojoin) > 1: # any keys specified
            autojoin_keys = dict(zip(autojoin[0].split(','), autojoin[1].split(',')))
            key = autojoin_keys.get(channel) # defaults to None when not set

    if key:
        w.command('', '/quote -server %s JOIN %s %s' % (server, channel, key))
    else:
        w.command('', '/quote -server %s JOIN %s' % (server, channel))

def invite_cb(data, signal, signal_data):
    server = signal.split(',')[0] # EFNet,irc_in_INVITE
    channel = signal_data.split()[-1].lstrip(':') # :nick!ident@host.name INVITE yournick :#channel
    from_nick = ''
    SearchStr = '(?:\@.*)?:(?P<nick>.+)!' #@time=2022-03-02T19:00:30.041Z :XX-XXXX!~XX-XXXX@xx.xxxx INVITE yournick :#xxxx-xxx (works also when no message-tag is present)
    from_nick = re.search(SearchStr, signal_data).groups()[0]

    if len(w.config_get_plugin('whitelist_nicks')) > 0 and len(w.config_get_plugin('whitelist_channels')) > 0: # if there's two whitelists, accept both
        if from_nick in w.config_get_plugin('whitelist_nicks').split(',') or channel in w.config_get_plugin('whitelist_channels').split(','):
            w.prnt('', 'Automatically joining %s on server %s, invitation from %s (whitelist).' \
                %(channel, server, from_nick))
            join(server, channel)
        else:
            w.prnt('', 'Ignoring invite from %s to channel %s. Neither inviter nor channel in whitelist.' %(from_nick, channel))

    elif len(w.config_get_plugin('whitelist_nicks')) > 0: # if there's a whitelist, accept nicks in it
        if from_nick in w.config_get_plugin('whitelist_nicks').split(','):
            w.prnt('', 'Automatically joining %s on server %s, invitation from %s (whitelist).' \
                %(channel, server, from_nick))
            join(server, channel)
        else:
            w.prnt('', 'Ignoring invite from %s to channel %s. Inviter not in whitelist.' %(from_nick, channel))

    elif len(w.config_get_plugin('whitelist_channels')) > 0: # if there's a whitelist, accept channels in it
        if channel in w.config_get_plugin('whitelist_channels').split(','):
            w.prnt('', 'Automatically joining %s on server %s, invitation from %s (whitelist).' \
                %(channel, server, from_nick))
            join(server, channel)
        else:
            w.prnt('', 'Ignoring invite from %s to channel %s. Channel not in whitelist.' %(from_nick, channel))

    else: # use the ignore lists to make the decision
        if from_nick in w.config_get_plugin('ignore_nicks').split(','):
            w.prnt('', 'Ignoring invite from %s to channel %s. Invite from ignored inviter.' %(from_nick, channel))
        elif channel in w.config_get_plugin('ignore_channels').split(','):
            w.prnt('', 'Ignoring invite from %s to channel %s. Invite to ignored channel.' %(from_nick, channel))
        else:
            w.prnt('', 'Automatically joining %s on server %s, invitation from %s.' \
                %(channel, server, from_nick))
            join(server, channel)

    return w.WEECHAT_RC_OK



import weechat as w

SCRIPT_NAME    = "noirccolors"
SCRIPT_AUTHOR  = "Fredrick Brennan <fredrick.brennan1@gmail.com>"
SCRIPT_VERSION = "0.4"
SCRIPT_LICENSE = "Public domain"
SCRIPT_DESC    = "Remove IRC colors from buffers with the localvar 'noirccolors' set. To disable IRC colors in the current buffer, type /buffer set localvar_set_noirccolors true. You can also set this with script buffer_autoset.py. :)"

w.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', '')


def my_modifier_cb(data, modifier, modifier_data, string):
    if modifier_data.startswith('0x'):
        # WeeChat >= 2.9
        buffer, tags = modifier_data.split(';', 1)
    else:
        # WeeChat <= 2.8
        plugin, buffer_name, tags = modifier_data.split(';', 2)
        buffer = w.buffer_search(plugin, buffer_name)
    if w.buffer_get_string(buffer, "localvar_noirccolors") == "true":
        try:
            nick, message = string.split("\t")
        except ValueError as e:
            return string
        return "%s\t%s" % (nick, w.string_remove_color(message,""))
    else:
        return string

hook = w.hook_modifier("weechat_print", "my_modifier_cb", "")



# ==================================================== #
# Script Name: cmdqueue.py
# Script Author: walk <KingPython@gmx.com>
# Script Purpose: Command queing at its finest. Hopefully.
#
# Copyright (C) 2011  walk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Version History:
# 0.4.5 - November 9th, 2021
#    Fix inconsistent use of tabs and spaces (now uses spaces as sugggested by PEP 8)
#    Improve formatting of code for better readability
#
# 0.4.4 - Sep 11th, 2021
#    Rename script to cmdqueue.py.
#
# 0.4.3 - May 5th, 2021
#    Add Python 3 compatibility.
#    Add compatibility with XDG directories (WeeChat >= 3.2).
#
# 0.4.2 - Nov 22nd, 2015
#    Add saving of static queues to disk and reloading them on startup.
#    Added by Tim Kuhlman - https://github.com/tkuhlman
#
# 0.4.1 - Jan 20th, 2011
#    Multi-list queuing seems to work flawlessly so far. Expanded on the /help qu text.
#    Properties are fully-functional. As for loading/saving of lists, I want to hold off until
#    I get some feedback on whether or not that would be applicable. Please email me
#    at the listed address in this script and let me know if you want this feature and/or
#    any other features.
#
# 0.4.0 - Jan 16th, 2011
#    Finished adding multi-list queuing. So far, no bugs that I can tell.
#    Perhaps an ability to load/save lists for multiple uses? Now working on
#    setting individual properties per list then taking a few days off this thing.
#    ** TESTED ON WeeChat 0.3.3 and 0.3.4 **
#
# 0.3.5 - Jan 15th, 2011
#    Started to implement multiple queue lists.
#
# 0.3.5 - Jan 14th, 2011
#    Wrote Queue class and merged with code. I did this because
#    some future release will include an ability to use multiple
#    queue lists. Also fixed some bugs in parsing and executing
#    arguments with /qu. /qu del by itself raised an error, /qu del j
#    raised a valueerror, fixed. /qu add singlearg wouldn't add. Fixed.
#
# 0.3.4 - Jan 11th, 2011
#    Modified configuration code to use plugins.conf
#
# 0.3.3 - Jan 9th, 2011
#    Big code clean-up. Reduced a few lines by optimizing code.
#    Next step, convertcommand_qu dictionary to array like
#    I should have done in the first place. Could save some lines.
#
# 0.3.2 - Jan 9th, 2011
#    Added RAINBOW option (requested)
#    Set as option in configuration
#
# 0.3.1 - Jan 8th, 2011
#    Added configurations (verbose, core output only)
#
# 0.3.0 - Jan 6th, 2011
#    Worked on script for quite a while. All code is functional.
#    Continuing to test for bugs and improve code
#    Fixed indexing issue after using /qu del with remove_index function
#
# 0.2.0 - Jan 5th, 2011
#   Cleaned up temporary testing code, built upon foundation.
#   Next to include, del function and list function
#
# 0.1.0 - Jan 4th, 2011
#   Wrote basic outline with minimal functionality
# ==================================================== #

import os
import pickle

import_ok = True
try:
    import weechat
except ImportError:
    print("This script requires WeeChat")
    print("To obtain a copy, for free, visit http://weechat.org")
    import_ok = False


class Queue():
    """ A Queuing Class """

    def __init__(self):
        self.data = []
        self.index = len(self.data)
        self.__clearable__ = True
        self.__ldl__ = ""
        self.__locked__ = False

    def __iter__(self):
        for char in range(self.index):
            yield self.data[char]

    def __len__(self):
        return len(self.data)

    def add(self, queue_text):
        if self.__locked__ == False:
            self.data.append(queue_text)
            self.index = len(self.data)

    def remove(self, info):
        if self.__locked__ == False:
            tmp = ""
            if info > 0:
                tmp = str(self.data[info-1])
                del self.data[info-1]
                self.__ldl__ = tmp
                self.index = len(self.data)
            elif info == 0:
                tmp = str(self.data[info])
                del self.data[info]
                self.__ldl__ = tmp
                self.index = len(self.data)

        return self.__ldl__

    def viewqueue(self):
        list = ""
        if not len(self.data) == 0:
            for each in range(len(self.data)):
                list += str(each+1) + ". " + str(self.data[each]) + "\n"
        else:
            list = "Nothing in queue"

        return list.strip("\n")

    def clearqueue(self):
        if self.__clearable__ == True:
            self.data = []
            self.index = 0

    def isClear(self, cOpt):
        if not cOpt in (True, False):
            return

        self.__clearable__ = cOpt

    def isLocked(self, lockOpt):
        if not lockOpt in (True, False):
            return

        self.__locked__ = lockOpt

    def isEmpty(self):
        if len(self.data) == 0:
            return True
        else:
            return False


SCRIPT_NAME = "cmdqueue"
SCRIPT_AUTHOR = "walk"
SCRIPT_VERSION = "0.4.5"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "Command queuing"

COMM_CMD = "qu"
COMM_DESC = "Queuing commands in WeeChat"
COMM_ARGS = "[add [command] | del [index] | new [list] | dellist [list] | set [property] [on|off] |list | clear | exec | listview]"
COMM_ARGS_DESC = "Examples: \n\
   /qu add /msg chanserv op #foo bar \n\
   /qu del 1 \n\
   /qu new weechat \n\
       - Use the 'new' argument to switch to already defined lists as well. \n\
   /qu dellist weechat \n\
   /qu list - List commands in current list \n\
   /qu list weechat - With optional parameter, you can choose to list the commands of a specified list. \n\
   /qu clear - Clear current list.. add a listname to clear a specified list. \n\
   /qu exec - Execute the commands of the current list.. you can also specify a list here as well. \n\
   /qu listview - Outputs the names of all your lists. \n\
   /qu save - Save static lists to disk \n\
   /qu set static on - Sets static property to ON for current list. This means that when executed, the list WILL NOT clear. The clear command will not work either.\n \
   \n\
   PROPERTIES (for set command):\n \
   static - prevents a list from clearing manually or automatically but can still add and del commands.\n \
   lock - prevents the user from adding/deleting entries to a list. Can be combined with static."
COMM_COMPL = "add|del|list|exec|new|listview|dellist"

COMMAND_QU = {"default": Queue()}
CURR_LIST = "default"


def __config__():
    """ Configuration initialization """

    if not weechat.config_get_plugin("core_output_only") in ("yes", "no"):
        weechat.config_set_plugin("core_output_only", "yes")
    if not weechat.config_get_plugin("rainbow_allow") in ("yes", "no"):
        weechat.config_set_plugin("rainbow_allow", "no")
    if not weechat.config_get_plugin("verbose") in ("yes", "no"):
        weechat.config_set_plugin("verbose", "yes")

    load()
    return weechat.WEECHAT_RC_OK


def load():
    """ Load saved queues from pickle. """
    global COMMAND_QU
    data_dir = weechat.info_get("weechat_data_dir", "") \
        or weechat.info_get("weechat_dir", "")
    pickle_path = os.path.join(data_dir, "queue.pickle")
    if os.path.exists(pickle_path):
        with open(pickle_path, "rb") as qu_pickle:
            COMMAND_QU = pickle.load(qu_pickle)

    if "default" not in COMMAND_QU:
        COMMAND_QU["default"] = Queue()


def rainbow(data):
    """ Not my favorite option but a requested one """

    colors = "red yellow green blue magenta"
    c = colors.split()
    count = 0
    colorHolder = ""

    for each in data:
        if count > 4:
            count = 0
        if not each == " ":
            colorHolder += weechat.color(c[count])+each
            count += 1
        else:
            colorHolder += " "

    return str(colorHolder)


def prntcore(data, essential=0, rb=0):
    """ Built more on weechat.prnt """

    if weechat.config_get_plugin("verbose") == "yes" or essential == 1:
        if weechat.config_get_plugin("core_output_only") == "yes":
            buffer = ""
        else:
            buffer = weechat.current_buffer()
        if rb == 0:
            weechat.prnt(buffer, data)
        else:
            weechat.prnt(buffer, rainbow(data))
    return weechat.WEECHAT_RC_OK


def save():
    """ Save to disk all static lists as a pickle. """
    global COMMAND_QU
    data_dir = weechat.info_get("weechat_data_dir", "") \
        or weechat.info_get("weechat_dir", "")
    pickle_path = os.path.join(data_dir, "queue.pickle")
    to_save = {}
    for name, qu in COMMAND_QU.items():
        if not qu.__clearable__:  # Note isClear method doesn't show status it sets it
            to_save[name] = qu

    with open(pickle_path, "wb") as qu_pickle:
        pickle.dump(to_save, qu_pickle, pickle.HIGHEST_PROTOCOL)


def rejoin(data, delimiter=" "):
    """ Rejoins a split string """
    tmpString = ""
    for each in data:
        tmpString += each+delimiter

    tmpString = tmpString.strip()
    return tmpString


def qu_cb(data, buffer, args):
    """ Process hook_command info """

    global CURR_LIST, COMMAND_QU
    if weechat.config_get_plugin("rainbow_allow") == "no":
        rainbowit = 0
    else:
        rainbowit = 1

    if args == "":
        return weechat.WEECHAT_RC_OK

    argv = args.split()
    arglist = ["add", "del", "new", "dellist", "list", "clear", "exec", "listview", "save", "set"]

    if not argv[0] in arglist:
        prntcore("[ queue -> not a valid argument: {0}".format(argv[0]), rb=rainbowit)
        return weechat.WEECHAT_RC_OK

    if argv[0].lower() == "add" and len(argv) > 1:
        if not COMMAND_QU[CURR_LIST].__locked__ == True:
            COMMAND_QU[CURR_LIST].add(rejoin(argv[1:]))
            prntcore("[ queue added -> "+str(rejoin(argv[1:])) + " ]", rb=rainbowit)
        else:
            prntcore("[ queue -> the lock property is enabled for this list ({0}). please disable it before adding/deleting. ]", rb=rainbowit)

    elif argv[0].lower() == "del" and len(argv) > 1:
        if not COMMAND_QU[CURR_LIST].__locked__ == True:
            try:
                rmd = COMMAND_QU[CURR_LIST].remove(int(argv[1]))
                prntcore("[ queue -> deleted: ({0}) {1} ]".format(argv[1], rmd), rb=rainbowit)
            except (IndexError, ValueError):
                prntcore("[ queue -> invalid reference. please check /qu list and try again. ]", rb=rainbowit)
        else:
            prntcore("[ queue -> the lock property is enabled for this list ({0}). please disable it before adding/deleting. ]".format(CURR_LIST), rb=rainbowit)

    elif argv[0].lower() == "clear":
        this_list = None
        if len(argv) > 1 and argv[1].lower() in COMMAND_QU.keys():
            this_list = CURR_LIST
            CURR_LIST = argv[1].lower()

        if COMMAND_QU[CURR_LIST].__clearable__ == True:
            if not COMMAND_QU[CURR_LIST].isEmpty():
                COMMAND_QU[CURR_LIST].clearqueue()
                prntcore("[ queue -> command queue list cleared. ]", rb=rainbowit)
            else:
                prntcore("[ queue -> command queue already empty. ]", rb=rainbowit)
        else:
            prntcore("[ queue -> please turn off the static property to clear the {0} list. ]".format(CURR_LIST), rb=rainbowit)

        if not this_list == None:
            CURR_LIST = this_list
            this_list = None

    elif argv[0].lower() == "list":
        this_list = None
        if len(argv) > 1 and argv[1].lower() in COMMAND_QU.keys():
            this_list = CURR_LIST
            CURR_LIST = argv[1].lower()

        qHeader = "[ COMMAND QUEUE: {0} ]".format(CURR_LIST)
        prntcore(" ", 1)
        prntcore("-"*len(qHeader), 1, rb=rainbowit)
        prntcore(qHeader, 1, rainbowit)
        prntcore("-"*len(qHeader), 1, rainbowit)
        prntcore(COMMAND_QU[CURR_LIST].viewqueue(), 1, rb=rainbowit)

        if not this_list == None:
            CURR_LIST = this_list
            this_list = None

    elif argv[0].lower() == "exec":

        this_list = None
        if len(argv) > 1 and argv[1].lower() in COMMAND_QU.keys():
            this_list = CURR_LIST
            CURR_LIST = argv[1].lower()

        if len(COMMAND_QU[CURR_LIST]) > 0:

            for each in COMMAND_QU[CURR_LIST]:
                weechat.command(buffer, each)
            COMMAND_QU[CURR_LIST].clearqueue()
            if COMMAND_QU[CURR_LIST].__clearable__ == True:
                prntcore("[ queue -> finished executing list: {0}. command list cleared. ]".format(CURR_LIST), rb=rainbowit)
            else:
                prntcore("[ queue -> finished executing list: {0} ]".format(CURR_LIST), rb=rainbowit)
        else:
            prntcore("[ queue -> nothing to execute. please add to the queue using /qu add <command>", rb=rainbowit)

        if not this_list == None:
            CURR_LIST = this_list
            this_list = None

    elif argv[0].lower() == "new" and len(args.split()) > 1:
        if argv[1].lower() in COMMAND_QU.keys():
            CURR_LIST = argv[1].lower()
            prntcore("[ queue -> switched queue list to: {0}".format(CURR_LIST), rb=rainbowit)
        else:
            COMMAND_QU[argv[1].lower()] = Queue()
            CURR_LIST = argv[1].lower()
            prntcore("[ queue -> created new list. current list is: {0}".format(CURR_LIST), rb=rainbowit)

    elif argv[0].lower() == "listview":
        qHeader = "QUEUE LISTS"
        listCount = 1

        prntcore(" ", 1)
        prntcore("-"*len(qHeader), 1, rb=rainbowit)
        prntcore(qHeader, 1, rb=rainbowit)
        prntcore("-"*len(qHeader), 1, rb=rainbowit)

        for each in COMMAND_QU.keys():
            prntcore(str(listCount) + ". " + str(each), 1, rb=rainbowit)
            listCount += 1

    elif argv[0].lower() == "dellist" and len(args.split()) > 1:
        if not argv[1].lower() in COMMAND_QU.keys():
            prntcore("[ queue -> {0} is not a list. ]".format(argv[1].lower()), rb=rainbowit)
        elif argv[1].lower() == "default":
            prntcore("[ queue -> cannot delete the default list. ]", rb=rainbowit)
        else:
            if argv[1].lower() == CURR_LIST:
                CURR_LIST = "default"
            del COMMAND_QU[argv[1].lower()]
            prntcore("[ queue -> {0} successfully deleted.".format(argv[1].lower()), rb=rainbowit)

    elif argv[0].lower() == "save":
        save()
    elif argv[0].lower() == "set" and len(argv) == 4:
        setargs = args.split()
        list_name = setargs[1].lower()
        set_prop = setargs[2].lower()
        toggle = setargs[3].lower()
        properties = ["static", "lock"]

        if not list_name in COMMAND_QU.keys():
            prntcore("[ queue -> list must be created before you can set properties ]", rb=rainbowit)
        elif not set_prop in properties:
            prntcore("[ queue -> invalid property. please try again. ]", rb=rainbowit)
        elif not toggle in ("on", "off"):
            prntcore("[ queue -> only valid options for a property are ON or OFF ]", rb=rainbowit)
        else:
            if set_prop == "static":
                if toggle == "on":
                    COMMAND_QU[list_name].isClear(False)
                    prntcore("[ queue -> static property toggled on for: {0} ]".format(list_name), rb=rainbowit)
                    save()
                else:
                    COMMAND_QU[list_name].isClear(True)
                    prntcore("[ queue -> static property toggled off for: {0} ]".format(list_name), rb=rainbowit)
                    save()

            elif set_prop == "lock":
                if toggle == "on":
                    COMMAND_QU[list_name].isLocked(True)
                    prntcore("[ queue -> lock property toggled on for: {0} ]".format(list_name), rb=rainbowit)
                else:
                    COMMAND_QU[list_name].isLocked(False)
                    prntcore("[ queue -> lock property toggled off for: {0} ]".format(list_name), rb=rainbowit)

    return weechat.WEECHAT_RC_OK


if import_ok and weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
    weechat.hook_command(COMM_CMD, COMM_DESC, COMM_ARGS, COMM_ARGS_DESC, COMM_COMPL, "qu_cb", "")
    __config__()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Wil Clouser <clouserw@micropipes.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Intercepts emoji characters and converts them back to their alias.  Useful if
# you're on a terminal or something that doesn't support emoji.
#
# History:
#
#   2019-07-06, Sébastien Helleu <flashcode@flashtux.org>:
#       v0.2: Make script compatible with Python 3, fix PEP8 errors.
#
#   2016-03-15, Wil Clouser <clouserw@micropipes.com>:
#       v0.1: Initial release, based on Mike Reinhardt's BSD
#             licensed emoji_aliases.py

from __future__ import print_function

import re
import sys

SCRIPT_NAME = "emoji2alias"
SCRIPT_AUTHOR = "Wil Clouser <clouserw@micropipes.com>"
SCRIPT_VERSION = "0.2"
SCRIPT_LICENSE = "MIT"
SCRIPT_DESC = "Replaces emoji characters with their aliases"

import_ok = True

try:
    import weechat as w
except ImportError:
    print("Script must be run under weechat: https://weechat.org")
    import_ok = False

EMOJI_ALIASES = {
    u'\U0001f44d': u':+1:',
    u'\U0001f44e': u':-1:',
    u'\U0001f4af': u':100:',
    u'\U0001f522': u':1234:',
    u'\U0001f3b1': u':8ball:',
    u'\U0001f170': u':a:',
    u'\U0001f18e': u':ab:',
    u'\U0001f524': u':abc:',
    u'\U0001f521': u':abcd:',
    u'\U0001f251': u':accept:',
    u'\U0001f39f': u':admission_tickets:',
    u'\U0001f6a1': u':aerial_tramway:',
    u'\U00002708': u':airplane:',
    u'\U0001f6ec': u':airplane_arriving:',
    u'\U0001f6eb': u':airplane_departure:',
    u'\U000023f0': u':alarm_clock:',
    u'\U00002697': u':alembic:',
    u'\U0001f47d': u':alien:',
    u'\U0001f47e': u':alien_monster:',
    u'\U0001f691': u':ambulance:',
    u'\U0001f3c8': u':american_football:',
    u'\U0001f3fa': u':amphora:',
    u'\U00002693': u':anchor:',
    u'\U0001f47c': u':angel:',
    u'\U0001f4a2': u':anger:',
    u'\U0001f620': u':angry:',
    u'\U0001f627': u':anguished:',
    u'\U0001f41c': u':ant:',
    u'\U0001f4f6': u':antenna_with_bars:',
    u'\U0001f34e': u':apple:',
    u'\U00002652': u':aquarius:',
    u'\U00002648': u':aries:',
    u'\U000025c0': u':arrow_backward:',
    u'\U000023ec': u':arrow_double_down:',
    u'\U000023eb': u':arrow_double_up:',
    u'\U00002b07': u':arrow_down:',
    u'\U0001f53d': u':arrow_down_small:',
    u'\U000025b6': u':arrow_forward:',
    u'\U00002935': u':arrow_heading_down:',
    u'\U00002934': u':arrow_heading_up:',
    u'\U00002b05': u':arrow_left:',
    u'\U00002199': u':arrow_lower_left:',
    u'\U00002198': u':arrow_lower_right:',
    u'\U000027a1': u':arrow_right:',
    u'\U000021aa': u':arrow_right_hook:',
    u'\U00002b06': u':arrow_up:',
    u'\U00002195': u':arrow_up_down:',
    u'\U0001f53c': u':arrow_up_small:',
    u'\U00002196': u':arrow_upper_left:',
    u'\U00002197': u':arrow_upper_right:',
    u'\U0001f503': u':arrows_clockwise:',
    u'\U0001f504': u':arrows_counterclockwise:',
    u'\U0001f3a8': u':art:',
    u'\U0001f69b': u':articulated_lorry:',
    u'\U0001f632': u':astonished:',
    u'\U0001f45f': u':athletic_shoe:',
    u'\U0001f3e7': u':atm:',
    u'\U0000269b': u':atom_symbol:',
    u'\U0001f346': u':aubergine:',
    u'\U0001f697': u':automobile:',
    u'\U0001f171': u':b:',
    u'\U0001f476': u':baby:',
    u'\U0001f37c': u':baby_bottle:',
    u'\U0001f424': u':baby_chick:',
    u'\U0001f6bc': u':baby_symbol:',
    u'\U0001f519': u':back:',
    u'\U0001f42b': u':bactrian_camel:',
    u'\U0001f3f8': u':badminton_racquet_and_shuttlecock:',
    u'\U0001f6c4': u':baggage_claim:',
    u'\U0001f388': u':balloon:',
    u'\U0001f5f3': u':ballot_box_with_ballot:',
    u'\U00002611': u':ballot_box_with_check:',
    u'\U0001f38d': u':bamboo:',
    u'\U0001f34c': u':banana:',
    u'\U0000203c': u':bangbang:',
    u'\U0001f3e6': u':bank:',
    u'\U0001f4b5': u':banknote_with_dollar_sign:',
    u'\U0001f4b6': u':banknote_with_euro_sign:',
    u'\U0001f4b7': u':banknote_with_pound_sign:',
    u'\U0001f4b4': u':banknote_with_yen_sign:',
    u'\U0001f4ca': u':bar_chart:',
    u'\U0001f488': u':barber:',
    u'\U0001f325': u':barely_sunny:',
    u'\U000026be': u':baseball:',
    u'\U0001f3c0': u':basketball:',
    u'\U0001f6c0': u':bath:',
    u'\U0001f6c1': u':bathtub:',
    u'\U0001f50b': u':battery:',
    u'\U0001f3d6': u':beach_with_umbrella:',
    u'\U0001f43b': u':bear:',
    u'\U0001f493': u':beating_heart:',
    u'\U0001f6cf': u':bed:',
    u'\U0001f41d': u':bee:',
    u'\U0001f37a': u':beer:',
    u'\U0001f37b': u':beers:',
    u'\U0001f41e': u':beetle:',
    u'\U0001f530': u':beginner:',
    u'\U0001f514': u':bell:',
    u'\U0001f515': u':bell_with_cancellation_stroke:',
    u'\U0001f6ce': u':bellhop_bell:',
    u'\U0001f371': u':bento:',
    u'\U0001f6b2': u':bicycle:',
    u'\U0001f6b4': u':bicyclist:',
    u'\U0001f459': u':bikini:',
    u'\U00002623': u':biohazard_sign:',
    u'\U0001f426': u':bird:',
    u'\U0001f382': u':birthday:',
    u'\U000026ab': u':black_circle:',
    u'\U000023fa': u':black_circle_for_record:',
    u'\U00002663': u':black_club_suit:',
    u'\U00002666': u':black_diamond_suit:',
    u'\U00002665': u':black_heart_suit:',
    u'\U0001f0cf': u':black_joker:',
    u'\U00002b1b': u':black_large_square:',
    u'\U000023ea': u':black_left-pointing_double_triangle:',
    u'\U000025fe': u':black_medium_small_square:',
    u'\U000025fc': u':black_medium_square:',
    u'\U00002712': u':black_nib:',
    u'\U00002753': u':black_question_mark_ornament:',
    u'\U000023e9': u':black_right-pointing_double_triangle:',
    u'\U00002702': u':black_scissors:',
    u'\U000025aa': u':black_small_square:',
    u'\U00002660': u':black_spade_suit:',
    u'\U0001f532': u':black_square_button:',
    u'\U000023f9': u':black_square_for_stop:',
    u'\U00002600': u':black_sun_with_rays:',
    u'\U0000260e': u':black_telephone:',
    u'\U0000267b': u':black_universal_recycling_symbol:',
    u'\U0001f33c': u':blossom:',
    u'\U0001f421': u':blowfish:',
    u'\U0001f4d8': u':blue_book:',
    u'\U0001f699': u':blue_car:',
    u'\U0001f499': u':blue_heart:',
    u'\U0001f60a': u':blush:',
    u'\U0001f417': u':boar:',
    u'\U000026f5': u':boat:',
    u'\U0001f4a3': u':bomb:',
    u'\U0001f4d6': u':book:',
    u'\U0001f516': u':bookmark:',
    u'\U0001f4d1': u':bookmark_tabs:',
    u'\U0001f4da': u':books:',
    u'\U0001f4a5': u':boom:',
    u'\U0001f462': u':boot:',
    u'\U0001f37e': u':bottle_with_popping_cork:',
    u'\U0001f490': u':bouquet:',
    u'\U0001f647': u':bow:',
    u'\U0001f3f9': u':bow_and_arrow:',
    u'\U0001f3b3': u':bowling:',
    u'\U0001f466': u':boy:',
    u'\U0001f35e': u':bread:',
    u'\U0001f470': u':bride_with_veil:',
    u'\U0001f309': u':bridge_at_night:',
    u'\U0001f4bc': u':briefcase:',
    u'\U0001f494': u':broken_heart:',
    u'\U0001f41b': u':bug:',
    u'\U0001f3d7': u':building_construction:',
    u'\U0001f4a1': u':bulb:',
    u'\U0001f685': u':bullettrain_front:',
    u'\U0001f684': u':bullettrain_side:',
    u'\U0001f32f': u':burrito:',
    u'\U0001f68c': u':bus:',
    u'\U0001f68f': u':bus_stop:',
    u'\U0001f464': u':bust_in_silhouette:',
    u'\U0001f465': u':busts_in_silhouette:',
    u'\U0001f335': u':cactus:',
    u'\U0001f370': u':cake:',
    u'\U0001f4c5': u':calendar:',
    u'\U0001f4c6': u':calendar:',
    u'\U0001f4f2': u':calling:',
    u'\U0001f4f7': u':camera:',
    u'\U0001f4f8': u':camera_with_flash:',
    u'\U0001f3d5': u':camping:',
    u'\U0000264b': u':cancer:',
    u'\U0001f56f': u':candle:',
    u'\U0001f36c': u':candy:',
    u'\U0001f520': u':capital_abcd:',
    u'\U00002651': u':capricorn:',
    u'\U0001f5c3': u':card_file_box:',
    u'\U0001f4c7': u':card_index:',
    u'\U0001f5c2': u':card_index_dividers:',
    u'\U0001f3a0': u':carousel_horse:',
    u'\U0001f38f': u':carp_streamer:',
    u'\U0001f408': u':cat2:',
    u'\U0001f431': u':cat:',
    u'\U0001f639': u':cat_face_with_tears_of_joy:',
    u'\U0001f63c': u':cat_face_with_wry_smile:',
    u'\U0001f4bf': u':cd:',
    u'\U000026d3': u':chains:',
    u'\U0001f4b9': u':chart:',
    u'\U0001f4c9': u':chart_with_downwards_trend:',
    u'\U0001f4c8': u':chart_with_upwards_trend:',
    u'\U0001f3c1': u':checkered_flag:',
    u'\U0001f4e3': u':cheering_megaphone:',
    u'\U0001f9c0': u':cheese_wedge:',
    u'\U0001f352': u':cherries:',
    u'\U0001f338': u':cherry_blossom:',
    u'\U0001f330': u':chestnut:',
    u'\U0001f414': u':chicken:',
    u'\U0001f6b8': u':children_crossing:',
    u'\U0001f43f': u':chipmunk:',
    u'\U0001f36b': u':chocolate_bar:',
    u'\U0001f384': u':christmas_tree:',
    u'\U000026ea': u':church:',
    u'\U0001f3a6': u':cinema:',
    u'\U0001f250': u':circled_ideograph_advantage:',
    u'\U00003297': u':circled_ideograph_congratulation:',
    u'\U00003299': u':circled_ideograph_secret:',
    u'\U000024c2': u':circled_latin_capital_letter_m:',
    u'\U0001f3aa': u':circus_tent:',
    u'\U0001f307': u':city_sunrise:',
    u'\U0001f306': u':city_sunset:',
    u'\U0001f3d9': u':cityscape:',
    u'\U0001f191': u':cl:',
    u'\U0001f44f': u':clap:',
    u'\U0001f3ac': u':clapper:',
    u'\U0001f3db': u':classical_building:',
    u'\U0001f4cb': u':clipboard:',
    u'\U0001f565': u':clock1030:',
    u'\U0001f559': u':clock10:',
    u'\U0001f566': u':clock1130:',
    u'\U0001f55a': u':clock11:',
    u'\U0001f567': u':clock1230:',
    u'\U0001f55b': u':clock12:',
    u'\U0001f55c': u':clock130:',
    u'\U0001f550': u':clock1:',
    u'\U0001f55d': u':clock230:',
    u'\U0001f551': u':clock2:',
    u'\U0001f55e': u':clock330:',
    u'\U0001f552': u':clock3:',
    u'\U0001f55f': u':clock430:',
    u'\U0001f553': u':clock4:',
    u'\U0001f560': u':clock530:',
    u'\U0001f554': u':clock5:',
    u'\U0001f561': u':clock630:',
    u'\U0001f555': u':clock6:',
    u'\U0001f562': u':clock730:',
    u'\U0001f556': u':clock7:',
    u'\U0001f563': u':clock830:',
    u'\U0001f557': u':clock8:',
    u'\U0001f564': u':clock930:',
    u'\U0001f558': u':clock9:',
    u'\U0001f4d5': u':closed_book:',
    u'\U0001f510': u':closed_lock_with_key:',
    u'\U0001f4ea': u':closed_mailbox_with_lowered_flag:',
    u'\U0001f4eb': u':closed_mailbox_with_raised_flag:',
    u'\U0001f302': u':closed_umbrella:',
    u'\U00002601': u':cloud:',
    u'\U0001f329': u':cloud_with_lightning:',
    u'\U0001f327': u':cloud_with_rain:',
    u'\U0001f328': u':cloud_with_snow:',
    u'\U0001f32a': u':cloud_with_tornado:',
    u'\U0001f378': u':cocktail:',
    u'\U00002615': u':coffee:',
    u'\U000026b0': u':coffin:',
    u'\U0001f630': u':cold_sweat:',
    u'\U00002604': u':comet:',
    u'\U0001f5dc': u':compression:',
    u'\U0001f4bb': u':computer:',
    u'\U0001f38a': u':confetti_ball:',
    u'\U0001f616': u':confounded:',
    u'\U0001f615': u':confused:',
    u'\U0001f6a7': u':construction:',
    u'\U0001f477': u':construction_worker:',
    u'\U0001f39b': u':control_knobs:',
    u'\U0001f3ea': u':convenience_store:',
    u'\U0001f35a': u':cooked_rice:',
    u'\U0001f36a': u':cookie:',
    u'\U0001f373': u':cooking:',
    u'\U0001f192': u':cool:',
    u'\U0001f46e': u':cop:',
    u'\U000000a9': u':copyright:',
    u'\U0001f33d': u':corn:',
    u'\U0001f6cb': u':couch_and_lamp:',
    u'\U0001f46b': u':couple:',
    u'\U0001f491': u':couple_with_heart:',
    u'\U0001f48f': u':couplekiss:',
    u'\U0001f404': u':cow2:',
    u'\U0001f42e': u':cow:',
    u'\U0001f980': u':crab:',
    u'\U0001f4b3': u':credit_card:',
    u'\U0001f319': u':crescent_moon:',
    u'\U0001f3cf': u':cricket_bat_and_ball:',
    u'\U0001f40a': u':crocodile:',
    u'\U0000274c': u':cross_mark:',
    u'\U0001f38c': u':crossed_flags:',
    u'\U00002694': u':crossed_swords:',
    u'\U0001f451': u':crown:',
    u'\U0001f622': u':cry:',
    u'\U0001f63f': u':crying_cat_face:',
    u'\U0001f52e': u':crystal_ball:',
    u'\U0001f498': u':cupid:',
    u'\U000027b0': u':curly_loop:',
    u'\U0001f4b1': u':currency_exchange:',
    u'\U0001f35b': u':curry:',
    u'\U0001f36e': u':custard:',
    u'\U0001f6c3': u':customs:',
    u'\U0001f300': u':cyclone:',
    u'\U0001f5e1': u':dagger_knife:',
    u'\U0001f483': u':dancer:',
    u'\U0001f46f': u':dancers:',
    u'\U0001f361': u':dango:',
    u'\U0001f576': u':dark_sunglasses:',
    u'\U0001f3af': u':dart:',
    u'\U0001f4a8': u':dash:',
    u'\U0001f333': u':deciduous_tree:',
    u'\U0001f69a': u':delivery_truck:',
    u'\U0001f3ec': u':department_store:',
    u'\U0001f3da': u':derelict_house_building:',
    u'\U0001f3dc': u':desert:',
    u'\U0001f3dd': u':desert_island:',
    u'\U0001f5a5': u':desktop_computer:',
    u'\U0001f4a0': u':diamond_shape_with_a_dot_inside:',
    u'\U0001f61e': u':disappointed:',
    u'\U0001f625': u':disappointed_but_relieved_face:',
    u'\U0001f4ab': u':dizzy:',
    u'\U0001f635': u':dizzy_face:',
    u'\U0001f6af': u':do_not_litter:',
    u'\U0001f415': u':dog2:',
    u'\U0001f436': u':dog:',
    u'\U0001f38e': u':dolls:',
    u'\U0001f42c': u':dolphin:',
    u'\U0001f6aa': u':door:',
    u'\U000027bf': u':double_curly_loop:',
    u'\U000023f8': u':double_vertical_bar:',
    u'\U0001f369': u':doughnut:',
    u'\U0001f54a': u':dove_of_peace:',
    u'\U0001f53b': u':down-pointing_red_triangle:',
    u'\U0001f409': u':dragon:',
    u'\U0001f432': u':dragon_face:',
    u'\U0001f457': u':dress:',
    u'\U0001f42a': u':dromedary_camel:',
    u'\U0001f4a7': u':droplet:',
    u'\U0001f4c0': u':dvd:',
    u'\U0001f4e7': u':e-mail:',
    u'\U0001f442': u':ear:',
    u'\U0001f33e': u':ear_of_rice:',
    u'\U0001f30d': u':earth_africa:',
    u'\U0001f30e': u':earth_americas:',
    u'\U0001f30f': u':earth_asia:',
    u'\U00002734': u':eight_pointed_black_star:',
    u'\U00002733': u':eight_spoked_asterisk:',
    u'\U000023cf': u':eject_symbol:',
    u'\U0001f50c': u':electric_plug:',
    u'\U0001f526': u':electric_torch:',
    u'\U0001f418': u':elephant:',
    u'\U00002709': u':email:',
    u'\U0001f3fb': u':emoji_modifier_fitzpatrick_type-1-2:',
    u'\U0001f3fc': u':emoji_modifier_fitzpatrick_type-3:',
    u'\U0001f3fd': u':emoji_modifier_fitzpatrick_type-4:',
    u'\U0001f3fe': u':emoji_modifier_fitzpatrick_type-5:',
    u'\U0001f3ff': u':emoji_modifier_fitzpatrick_type-6:',
    u'\U0001f51a': u':end:',
    u'\U0001f4e9': u':envelope_with_arrow:',
    u'\U0001f3f0': u':european_castle:',
    u'\U0001f3e4': u':european_post_office:',
    u'\U0001f332': u':evergreen_tree:',
    u'\U00002757': u':exclamation:',
    u'\U00002049': u':exclamation_question_mark:',
    u'\U0001f611': u':expressionless:',
    u'\U0001f441': u':eye:',
    u'\U0001f453': u':eyeglasses:',
    u'\U0001f440': u':eyes:',
    u'\U0001f486': u':face_massage:',
    u'\U0001f60b': u':face_savouring_delicious_food:',
    u'\U0001f631': u':face_screaming_in_fear:',
    u'\U0001f618': u':face_throwing_a_kiss:',
    u'\U0001f613': u':face_with_cold_sweat:',
    u'\U0001f915': u':face_with_head-bandage:',
    u'\U0001f624': u':face_with_look_of_triumph:',
    u'\U0001f637': u':face_with_medical_mask:',
    u'\U0001f645': u':face_with_no_good_gesture:',
    u'\U0001f646': u':face_with_ok_gesture:',
    u'\U0001f62e': u':face_with_open_mouth:',
    u'\U0001f644': u':face_with_rolling_eyes:',
    u'\U0001f61b': u':face_with_stuck-out_tongue:',
    u'\U0001f602': u':face_with_tears_of_joy:',
    u'\U0001f912': u':face_with_thermometer:',
    u'\U0001f636': u':face_without_mouth:',
    u'\U0001f44a': u':facepunch:',
    u'\U0001f3ed': u':factory:',
    u'\U0001f342': u':fallen_leaf:',
    u'\U0001f46a': u':family:',
    u'\U0001f385': u':father_christmas:',
    u'\U0001f4e0': u':fax:',
    u'\U0001f628': u':fearful:',
    u'\U0001f43e': u':feet:',
    u'\U0001f3a1': u':ferris_wheel:',
    u'\U000026f4': u':ferry:',
    u'\U0001f3d1': u':field_hockey_stick_and_ball:',
    u'\U0001f5c4': u':file_cabinet:',
    u'\U0001f4c1': u':file_folder:',
    u'\U0001f39e': u':film_frames:',
    u'\U0001f4fd': u':film_projector:',
    u'\U0001f525': u':fire:',
    u'\U0001f692': u':fire_engine:',
    u'\U0001f387': u':firework_sparkler:',
    u'\U0001f386': u':fireworks:',
    u'\U0001f313': u':first_quarter_moon:',
    u'\U0001f31b': u':first_quarter_moon_with_face:',
    u'\U0001f41f': u':fish:',
    u'\U0001f365': u':fish_cake:',
    u'\U0001f3a3': u':fishing_pole_and_fish:',
    u'\U0000270a': u':fist:',
    u'\U000026f3': u':flag_in_hole:',
    u'\U0000269c': u':fleur-de-lis:',
    u'\U0001f4aa': u':flexed_biceps:',
    u'\U0001f4be': u':floppy_disk:',
    u'\U0001f3b4': u':flower_playing_cards:',
    u'\U0001f633': u':flushed:',
    u'\U0001f32b': u':fog:',
    u'\U0001f301': u':foggy:',
    u'\U0001f463': u':footprints:',
    u'\U0001f374': u':fork_and_knife:',
    u'\U0001f37d': u':fork_and_knife_with_plate:',
    u'\U000026f2': u':fountain:',
    u'\U0001f340': u':four_leaf_clover:',
    u'\U0001f5bc': u':frame_with_picture:',
    u'\U0001f193': u':free:',
    u'\U0001f35f': u':french_fries:',
    u'\U0001f364': u':fried_shrimp:',
    u'\U0001f438': u':frog:',
    u'\U0001f425': u':front-facing_baby_chick:',
    u'\U0001f626': u':frowning:',
    u'\U000026fd': u':fuel_pump:',
    u'\U0001f315': u':full_moon:',
    u'\U0001f31d': u':full_moon_with_face:',
    u'\U000026b1': u':funeral_urn:',
    u'\U0001f3b2': u':game_die:',
    u'\U00002699': u':gear:',
    u'\U0001f48e': u':gem:',
    u'\U0000264a': u':gemini:',
    u'\U0001f47b': u':ghost:',
    u'\U0001f381': u':gift:',
    u'\U0001f49d': u':gift_heart:',
    u'\U0001f467': u':girl:',
    u'\U0001f310': u':globe_with_meridians:',
    u'\U0001f31f': u':glowing_star:',
    u'\U0001f410': u':goat:',
    u'\U0001f3cc': u':golfer:',
    u'\U0001f393': u':graduation_cap:',
    u'\U0001f347': u':grapes:',
    u'\U0001f34f': u':green_apple:',
    u'\U0001f4d7': u':green_book:',
    u'\U0001f49a': u':green_heart:',
    u'\U00002755': u':grey_exclamation:',
    u'\U00002754': u':grey_question:',
    u'\U0001f62c': u':grimacing:',
    u'\U0001f601': u':grin:',
    u'\U0001f600': u':grinning:',
    u'\U0001f638': u':grinning_cat_face_with_smiling_eyes:',
    u'\U0001f497': u':growing_heart:',
    u'\U0001f482': u':guardsman:',
    u'\U0001f3b8': u':guitar:',
    u'\U0001f52b': u':gun:',
    u'\U0001f487': u':haircut:',
    u'\U0001f354': u':hamburger:',
    u'\U0001f528': u':hammer:',
    u'\U00002692': u':hammer_and_pick:',
    u'\U0001f6e0': u':hammer_and_wrench:',
    u'\U0001f439': u':hamster:',
    u'\U0000270b': u':hand:',
    u'\U0001f45c': u':handbag:',
    u'\U0001f4a9': u':hankey:',
    u'\U0001f64b': u':happy_person_raising_one_hand:',
    u'\U0001f423': u':hatching_chick:',
    u'\U0001f3a7': u':headphone:',
    u'\U0001f649': u':hear-no-evil_monkey:',
    u'\U00002764': u':heart:',
    u'\U0001f49f': u':heart_decoration:',
    u'\U0001f60d': u':heart_eyes:',
    u'\U0001f63b': u':heart_eyes_cat:',
    u'\U00002714': u':heavy_check_mark:',
    u'\U00002797': u':heavy_division_sign:',
    u'\U0001f4b2': u':heavy_dollar_sign:',
    u'\U00002763': u':heavy_heart_exclamation_mark_ornament:',
    u'\U00002b55': u':heavy_large_circle:',
    u'\U00002796': u':heavy_minus_sign:',
    u'\U00002716': u':heavy_multiplication_x:',
    u'\U00002795': u':heavy_plus_sign:',
    u'\U0001f681': u':helicopter:',
    u'\U00002388': u':helm_symbol:',
    u'\U000026d1': u':helmet_with_white_cross:',
    u'\U0001f33f': u':herb:',
    u'\U0001f33a': u':hibiscus:',
    u'\U0001f460': u':high-heeled_shoe:',
    u'\U0001f506': u':high_brightness:',
    u'\U000026a1': u':high_voltage_sign:',
    u'\U0001f52a': u':hocho:',
    u'\U0001f573': u':hole:',
    u'\U0001f36f': u':honey_pot:',
    u'\U0001f6a5': u':horizontal_traffic_light:',
    u'\U0001f40e': u':horse:',
    u'\U0001f434': u':horse:',
    u'\U0001f3c7': u':horse_racing:',
    u'\U0001f3e5': u':hospital:',
    u'\U0001f32d': u':hot_dog:',
    u'\U0001f336': u':hot_pepper:',
    u'\U00002668': u':hot_springs:',
    u'\U0001f3e8': u':hotel:',
    u'\U0000231b': u':hourglass:',
    u'\U000023f3': u':hourglass_flowing_sand:',
    u'\U0001f3e0': u':house:',
    u'\U0001f3d8': u':house_buildings:',
    u'\U0001f3e1': u':house_with_garden:',
    u'\U0001f917': u':hugging_face:',
    u'\U0001f62f': u':hushed:',
    u'\U0001f368': u':ice_cream:',
    u'\U0001f3d2': u':ice_hockey_stick_and_puck:',
    u'\U000026f8': u':ice_skate:',
    u'\U0001f366': u':icecream:',
    u'\U0001f194': u':id:',
    u'\U0001f47f': u':imp:',
    u'\U0001f4e5': u':inbox_tray:',
    u'\U0001f4e8': u':incoming_envelope:',
    u'\U0001f481': u':information_desk_person:',
    u'\U00002139': u':information_source:',
    u'\U0001f607': u':innocent:',
    u'\U0001f523': u':input_symbol_for_symbols:',
    u'\U0001f4f1': u':iphone:',
    u'\U0001f3ee': u':izakaya_lantern:',
    u'\U0001f383': u':jack-o-lantern:',
    u'\U0001f5fe': u':japan:',
    u'\U0001f3ef': u':japanese_castle:',
    u'\U0001f47a': u':japanese_goblin:',
    u'\U0001f479': u':japanese_ogre:',
    u'\U0001f3e3': u':japanese_post_office:',
    u'\U0001f456': u':jeans:',
    u'\U0001f579': u':joystick:',
    u'\U0001f54b': u':kaaba:',
    u'\U0001f511': u':key:',
    u'\U00002328': u':keyboard:',
    u'\U0001f51f': u':keycap_ten:',
    u'\U0001f458': u':kimono:',
    u'\U0001f48b': u':kiss:',
    u'\U0001f617': u':kissing:',
    u'\U0001f63d': u':kissing_cat:',
    u'\U0001f61a': u':kissing_closed_eyes:',
    u'\U0001f619': u':kissing_face_with_smiling_eyes:',
    u'\U0001f428': u':koala:',
    u'\U0001f201': u':koko:',
    u'\U0001f3f7': u':label:',
    u'\U0001f535': u':large_blue_circle:',
    u'\U0001f537': u':large_blue_diamond:',
    u'\U0001f536': u':large_orange_diamond:',
    u'\U0001f534': u':large_red_circle:',
    u'\U0001f317': u':last_quarter_moon:',
    u'\U0001f31c': u':last_quarter_moon_with_face:',
    u'\U0000271d': u':latin_cross:',
    u'\U0001f606': u':laughing:',
    u'\U0001f343': u':leaf_fluttering_in_wind:',
    u'\U0001f4d2': u':ledger:',
    u'\U0001f50d': u':left-pointing_magnifying_glass:',
    u'\U0001f6c5': u':left_luggage:',
    u'\U00002194': u':left_right_arrow:',
    u'\U0001f4ac': u':left_speech_bubble:',
    u'\U000021a9': u':leftwards_arrow_with_hook:',
    u'\U0001f34b': u':lemon:',
    u'\U0000264c': u':leo:',
    u'\U0001f406': u':leopard:',
    u'\U0001f39a': u':level_slider:',
    u'\U0000264e': u':libra:',
    u'\U0001f688': u':light_rail:',
    u'\U0001f517': u':link:',
    u'\U0001f587': u':linked_paperclips:',
    u'\U0001f981': u':lion_face:',
    u'\U0001f444': u':lips:',
    u'\U0001f484': u':lipstick:',
    u'\U0001f512': u':lock:',
    u'\U0001f50f': u':lock_with_ink_pen:',
    u'\U0001f36d': u':lollipop:',
    u'\U0001f50a': u':loud_sound:',
    u'\U0001f62d': u':loudly_crying_face:',
    u'\U0001f4e2': u':loudspeaker:',
    u'\U0001f3e9': u':love_hotel:',
    u'\U0001f48c': u':love_letter:',
    u'\U0001f505': u':low_brightness:',
    u'\U0001f58a': u':lower_left_ballpoint_pen:',
    u'\U0001f58d': u':lower_left_crayon:',
    u'\U0001f58b': u':lower_left_fountain_pen:',
    u'\U0001f58c': u':lower_left_paintbrush:',
    u'\U0001f50e': u':mag_right:',
    u'\U0001f004': u':mahjong:',
    u'\U0001f4ec': u':mailbox_with_mail:',
    u'\U0001f4ed': u':mailbox_with_no_mail:',
    u'\U0001f468': u':man:',
    u'\U0001f574': u':man_in_business_suit_levitating:',
    u'\U0001f472': u':man_with_gua_pi_mao:',
    u'\U0001f473': u':man_with_turban:',
    u'\U0001f45e': u':mans_shoe:',
    u'\U0001f570': u':mantelpiece_clock:',
    u'\U0001f341': u':maple_leaf:',
    u'\U0001f356': u':meat_on_bone:',
    u'\U0001f3c5': u':medal:',
    u'\U000026aa': u':medium_white_circle:',
    u'\U0001f348': u':melon:',
    u'\U0001f4dd': u':memo:',
    u'\U0001f54e': u':menorah_with_nine_branches:',
    u'\U0001f6b9': u':mens:',
    u'\U0001f687': u':metro:',
    u'\U0001f3a4': u':microphone:',
    u'\U0001f52c': u':microscope:',
    u'\U0001f595': u':middle_finger:',
    u'\U0001f396': u':military_medal:',
    u'\U0001f30c': u':milky_way:',
    u'\U0001f690': u':minibus:',
    u'\U0001f4bd': u':minidisc:',
    u'\U0001f4f4': u':mobile_phone_off:',
    u'\U0001f911': u':money-mouth_face:',
    u'\U0001f4b0': u':money_bag:',
    u'\U0001f4b8': u':money_with_wings:',
    u'\U0001f412': u':monkey:',
    u'\U0001f435': u':monkey_face:',
    u'\U0001f69d': u':monorail:',
    u'\U0001f314': u':moon:',
    u'\U0001f391': u':moon_viewing_ceremony:',
    u'\U0001f54c': u':mosque:',
    u'\U0001f324': u':mostly_sunny:',
    u'\U0001f6e5': u':motor_boat:',
    u'\U0001f6e3': u':motorway:',
    u'\U0001f5fb': u':mount_fuji:',
    u'\U000026f0': u':mountain:',
    u'\U0001f6b5': u':mountain_bicyclist:',
    u'\U0001f6a0': u':mountain_cableway:',
    u'\U0001f69e': u':mountain_railway:',
    u'\U0001f401': u':mouse2:',
    u'\U0001f42d': u':mouse:',
    u'\U0001f3a5': u':movie_camera:',
    u'\U0001f5ff': u':moyai:',
    u'\U0001f3b6': u':multiple_musical_notes:',
    u'\U0001f344': u':mushroom:',
    u'\U0001f3b9': u':musical_keyboard:',
    u'\U0001f3b5': u':musical_note:',
    u'\U0001f3bc': u':musical_score:',
    u'\U0001f507': u':mute:',
    u'\U0001f485': u':nail_care:',
    u'\U0001f4db': u':name_badge:',
    u'\U0001f3de': u':national_park:',
    u'\U0001f454': u':necktie:',
    u'\U0000274e': u':negative_squared_cross_mark:',
    u'\U0001f913': u':nerd_face:',
    u'\U0001f610': u':neutral_face:',
    u'\U0001f195': u':new:',
    u'\U0001f311': u':new_moon:',
    u'\U0001f31a': u':new_moon_with_face:',
    u'\U0001f4f0': u':newspaper:',
    u'\U0001f196': u':ng:',
    u'\U0001f303': u':night_with_stars:',
    u'\U0001f6b3': u':no_bicycles:',
    u'\U000026d4': u':no_entry:',
    u'\U0001f6ab': u':no_entry_sign:',
    u'\U0001f4f5': u':no_mobile_phones:',
    u'\U0001f51e': u':no_one_under_eighteen_symbol:',
    u'\U0001f6b7': u':no_pedestrians:',
    u'\U0001f6ad': u':no_smoking:',
    u'\U0001f6b1': u':non-potable_water:',
    u'\U0001f443': u':nose:',
    u'\U0001f4d3': u':notebook:',
    u'\U0001f4d4': u':notebook_with_decorative_cover:',
    u'\U0001f529': u':nut_and_bolt:',
    u'\U0001f17e': u':o2:',
    u'\U0001f30a': u':ocean:',
    u'\U0001f419': u':octopus:',
    u'\U0001f362': u':oden:',
    u'\U0001f3e2': u':office:',
    u'\U0001f6e2': u':oil_drum:',
    u'\U0001f197': u':ok:',
    u'\U0001f44c': u':ok_hand:',
    u'\U0001f5dd': u':old_key:',
    u'\U0001f474': u':older_man:',
    u'\U0001f475': u':older_woman:',
    u'\U0001f549': u':om_symbol:',
    u'\U0001f51b': u':on:',
    u'\U0001f698': u':oncoming_automobile:',
    u'\U0001f68d': u':oncoming_bus:',
    u'\U0001f694': u':oncoming_police_car:',
    u'\U0001f696': u':oncoming_taxi:',
    u'\U0001f4c2': u':open_file_folder:',
    u'\U0001f450': u':open_hands:',
    u'\U0001f513': u':open_lock:',
    u'\U000026ce': u':ophiuchus:',
    u'\U0001f4d9': u':orange_book:',
    u'\U00002626': u':orthodox_cross:',
    u'\U0001f4e4': u':outbox_tray:',
    u'\U0001f402': u':ox:',
    u'\U0001f4e6': u':package:',
    u'\U0001f4c4': u':page_facing_up:',
    u'\U0001f4c3': u':page_with_curl:',
    u'\U0001f4df': u':pager:',
    u'\U0001f334': u':palm_tree:',
    u'\U0001f43c': u':panda_face:',
    u'\U0001f4ce': u':paperclip:',
    u'\U0001f17f': u':parking:',
    u'\U0000303d': u':part_alternation_mark:',
    u'\U000026c5': u':partly_sunny:',
    u'\U0001f326': u':partly_sunny_rain:',
    u'\U0001f389': u':party_popper:',
    u'\U0001f6f3': u':passenger_ship:',
    u'\U0001f6c2': u':passport_control:',
    u'\U0000262e': u':peace_symbol:',
    u'\U0001f351': u':peach:',
    u'\U0001f350': u':pear:',
    u'\U0001f6b6': u':pedestrian:',
    u'\U0000270f': u':pencil2:',
    u'\U0001f427': u':penguin:',
    u'\U0001f614': u':pensive:',
    u'\U0001f3ad': u':performing_arts:',
    u'\U0001f623': u':persevere:',
    u'\U0001f64d': u':person_frowning:',
    u'\U000026f9': u':person_with_ball:',
    u'\U0001f471': u':person_with_blond_hair:',
    u'\U0001f64f': u':person_with_folded_hands:',
    u'\U0001f64e': u':person_with_pouting_face:',
    u'\U000026cf': u':pick:',
    u'\U0001f416': u':pig2:',
    u'\U0001f437': u':pig:',
    u'\U0001f43d': u':pig_nose:',
    u'\U0001f48a': u':pill:',
    u'\U0001f34d': u':pineapple:',
    u'\U00002653': u':pisces:',
    u'\U0001f355': u':pizza:',
    u'\U0001f6d0': u':place_of_worship:',
    u'\U0001f447': u':point_down:',
    u'\U0001f448': u':point_left:',
    u'\U0001f449': u':point_right:',
    u'\U0000261d': u':point_up:',
    u'\U0001f446': u':point_up_2:',
    u'\U0001f693': u':police_car:',
    u'\U0001f6a8': u':police_cars_revolving_light:',
    u'\U0001f429': u':poodle:',
    u'\U0001f37f': u':popcorn:',
    u'\U0001f4ef': u':postal_horn:',
    u'\U0001f4ee': u':postbox:',
    u'\U0001f372': u':pot_of_food:',
    u'\U0001f6b0': u':potable_water:',
    u'\U0001f45d': u':pouch:',
    u'\U0001f357': u':poultry_leg:',
    u'\U0001f63e': u':pouting_cat:',
    u'\U0001f621': u':pouting_face:',
    u'\U0001f4ff': u':prayer_beads:',
    u'\U0001f478': u':princess:',
    u'\U0001f5a8': u':printer:',
    u'\U0001f49c': u':purple_heart:',
    u'\U0001f45b': u':purse:',
    u'\U0001f4cc': u':pushpin:',
    u'\U0001f6ae': u':put_litter_in_its_place:',
    u'\U0001f407': u':rabbit2:',
    u'\U0001f430': u':rabbit:',
    u'\U0001f3ce': u':racing_car:',
    u'\U0001f3cd': u':racing_motorcycle:',
    u'\U0001f4fb': u':radio:',
    u'\U0001f518': u':radio_button:',
    u'\U00002622': u':radioactive_sign:',
    u'\U0001f683': u':railway_car:',
    u'\U0001f6e4': u':railway_track:',
    u'\U0001f308': u':rainbow:',
    u'\U0001f590': u':raised_hand_with_fingers_splayed:',
    u'\U0001f64c': u':raised_hands:',
    u'\U0001f40f': u':ram:',
    u'\U0001f35c': u':ramen:',
    u'\U0001f400': u':rat:',
    u'\U000000ae': u':registered:',
    u'\U0000263a': u':relaxed:',
    u'\U0001f60c': u':relieved:',
    u'\U0001f397': u':reminder_ribbon:',
    u'\U0001f501': u':repeat:',
    u'\U0001f502': u':repeat_one:',
    u'\U0001f6bb': u':restroom:',
    u'\U0001f49e': u':revolving_hearts:',
    u'\U0001f380': u':ribbon:',
    u'\U0001f359': u':rice_ball:',
    u'\U0001f358': u':rice_cracker:',
    u'\U0001f5ef': u':right_anger_bubble:',
    u'\U0001f48d': u':ring:',
    u'\U0001f360': u':roasted_sweet_potato:',
    u'\U0001f916': u':robot_face:',
    u'\U0001f680': u':rocket:',
    u'\U0001f5de': u':rolled-up_newspaper:',
    u'\U0001f3a2': u':roller_coaster:',
    u'\U0001f413': u':rooster:',
    u'\U0001f339': u':rose:',
    u'\U0001f3f5': u':rosette:',
    u'\U0001f4cd': u':round_pushpin:',
    u'\U0001f6a3': u':rowboat:',
    u'\U0001f3c9': u':rugby_football:',
    u'\U0001f3c3': u':runner:',
    u'\U0001f3bd': u':running_shirt_with_sash:',
    u'\U0001f202': u':sa:',
    u'\U00002650': u':sagittarius:',
    u'\U0001f376': u':sake:',
    u'\U0001f461': u':sandal:',
    u'\U0001f4e1': u':satellite:',
    u'\U0001f6f0': u':satellite:',
    u'\U0001f3b7': u':saxophone:',
    u'\U00002696': u':scales:',
    u'\U0001f3eb': u':school:',
    u'\U0001f392': u':school_satchel:',
    u'\U0001f982': u':scorpion:',
    u'\U0000264f': u':scorpius:',
    u'\U0001f640': u':scream_cat:',
    u'\U0001f4dc': u':scroll:',
    u'\U0001f4ba': u':seat:',
    u'\U0001f648': u':see-no-evil_monkey:',
    u'\U0001f331': u':seedling:',
    u'\U00002618': u':shamrock:',
    u'\U0001f367': u':shaved_ice:',
    u'\U0001f411': u':sheep:',
    u'\U0001f41a': u':shell:',
    u'\U0001f6e1': u':shield:',
    u'\U000026e9': u':shinto_shrine:',
    u'\U0001f6a2': u':ship:',
    u'\U0001f455': u':shirt:',
    u'\U0001f320': u':shooting_star:',
    u'\U0001f6cd': u':shopping_bags:',
    u'\U0001f6bf': u':shower:',
    u'\U0001f918': u':sign_of_the_horns:',
    u'\U0001f642': u':simple_smile:',
    u'\U0001f52f': u':six_pointed_star:',
    u'\U0001f3bf': u':ski:',
    u'\U000026f7': u':skier:',
    u'\U0001f480': u':skull:',
    u'\U00002620': u':skull_and_crossbones:',
    u'\U0001f634': u':sleeping:',
    u'\U0001f6cc': u':sleeping_accommodation:',
    u'\U0001f4a4': u':sleeping_symbol:',
    u'\U0001f62a': u':sleepy:',
    u'\U0001f575': u':sleuth_or_spy:',
    u'\U0001f641': u':slightly_frowning_face:',
    u'\U0001f3b0': u':slot_machine:',
    u'\U0001f6e9': u':small_airplane:',
    u'\U0001f539': u':small_blue_diamond:',
    u'\U0001f538': u':small_orange_diamond:',
    u'\U0001f53a': u':small_red_triangle:',
    u'\U0001f604': u':smile:',
    u'\U0001f603': u':smiley:',
    u'\U0001f63a': u':smiley_cat:',
    u'\U0001f608': u':smiling_face_with_horns:',
    u'\U0001f605': u':smiling_face_with_open_mouth_and_cold_sweat:',
    u'\U0001f60e': u':smiling_face_with_sunglasses:',
    u'\U0001f60f': u':smirk:',
    u'\U0001f6ac': u':smoking:',
    u'\U0001f40c': u':snail:',
    u'\U0001f40d': u':snake:',
    u'\U0001f3d4': u':snow_capped_mountain:',
    u'\U0001f3c2': u':snowboarder:',
    u'\U00002744': u':snowflake:',
    u'\U00002603': u':snowman:',
    u'\U000026c4': u':snowman_without_snow:',
    u'\U000026bd': u':soccer:',
    u'\U0001f51c': u':soon:',
    u'\U0001f198': u':sos:',
    u'\U0001f509': u':sound:',
    u'\U0001f35d': u':spaghetti:',
    u'\U00002747': u':sparkle:',
    u'\U00002728': u':sparkles:',
    u'\U0001f496': u':sparkling_heart:',
    u'\U0001f64a': u':speak-no-evil_monkey:',
    u'\U0001f508': u':speaker:',
    u'\U0001f5e3': u':speaking_head_in_silhouette:',
    u'\U0001f6a4': u':speedboat:',
    u'\U0001f577': u':spider:',
    u'\U0001f578': u':spider_web:',
    u'\U0001f5d3': u':spiral_calendar_pad:',
    u'\U0001f5d2': u':spiral_note_pad:',
    u'\U0001f4a6': u':splashing_sweat_symbol:',
    u'\U0001f596': u':spock-hand:',
    u'\U0001f433': u':spouting_whale:',
    u'\U0001f199': u':squared_up_with_exclamation_mark:',
    u'\U0001f19a': u':squared_vs:',
    u'\U0001f3df': u':stadium:',
    u'\U00002b50': u':star:',
    u'\U0000262a': u':star_and_crescent:',
    u'\U00002721': u':star_of_david:',
    u'\U0001f689': u':station:',
    u'\U0001f5fd': u':statue_of_liberty:',
    u'\U0001f682': u':steam_locomotive:',
    u'\U000023f1': u':stopwatch:',
    u'\U0001f4cf': u':straight_ruler:',
    u'\U0001f353': u':strawberry:',
    u'\U0001f61d': u':stuck_out_tongue_closed_eyes:',
    u'\U0001f61c': u':stuck_out_tongue_winking_eye:',
    u'\U0001f399': u':studio_microphone:',
    u'\U0001f31e': u':sun_with_face:',
    u'\U0001f33b': u':sunflower:',
    u'\U0001f305': u':sunrise:',
    u'\U0001f304': u':sunrise_over_mountains:',
    u'\U0001f3c4': u':surfer:',
    u'\U0001f363': u':sushi:',
    u'\U0001f69f': u':suspension_railway:',
    u'\U0001f3ca': u':swimmer:',
    u'\U0001f54d': u':synagogue:',
    u'\U0001f489': u':syringe:',
    u'\U0001f3d3': u':table_tennis_paddle_and_ball:',
    u'\U0001f32e': u':taco:',
    u'\U0001f38b': u':tanabata_tree:',
    u'\U0001f34a': u':tangerine:',
    u'\U00002649': u':taurus:',
    u'\U0001f695': u':taxi:',
    u'\U0001f375': u':tea:',
    u'\U0001f4de': u':telephone_receiver:',
    u'\U0001f52d': u':telescope:',
    u'\U0001f4fa': u':television:',
    u'\U0001f3be': u':tennis:',
    u'\U000026fa': u':tent:',
    u'\U0001f321': u':thermometer:',
    u'\U0001f914': u':thinking_face:',
    u'\U0001f4ad': u':thought_balloon:',
    u'\U0001f5b1': u':three_button_mouse:',
    u'\U000026c8': u':thunder_cloud_and_rain:',
    u'\U0001f3ab': u':ticket:',
    u'\U0001f405': u':tiger2:',
    u'\U0001f42f': u':tiger:',
    u'\U000023f2': u':timer_clock:',
    u'\U0001f62b': u':tired_face:',
    u'\U00002122': u':tm:',
    u'\U0001f6bd': u':toilet:',
    u'\U0001f5fc': u':tokyo_tower:',
    u'\U0001f345': u':tomato:',
    u'\U0001f445': u':tongue:',
    u'\U0001f51d': u':top:',
    u'\U0001f3a9': u':top_hat:',
    u'\U0001f5b2': u':trackball:',
    u'\U0001f69c': u':tractor:',
    u'\U0001f686': u':train2:',
    u'\U0001f68b': u':train:',
    u'\U0001f68a': u':tram:',
    u'\U0001f6a9': u':triangular_flag_on_post:',
    u'\U0001f4d0': u':triangular_ruler:',
    u'\U0001f531': u':trident:',
    u'\U0001f68e': u':trolleybus:',
    u'\U0001f3c6': u':trophy:',
    u'\U0001f379': u':tropical_drink:',
    u'\U0001f420': u':tropical_fish:',
    u'\U0001f3ba': u':trumpet:',
    u'\U0001f337': u':tulip:',
    u'\U0001f983': u':turkey:',
    u'\U0001f422': u':turtle:',
    u'\U0001f500': u':twisted_rightwards_arrows:',
    u'\U0001f495': u':two_hearts:',
    u'\U0001f46c': u':two_men_holding_hands:',
    u'\U0001f46d': u':two_women_holding_hands:',
    u'\U00002602': u':umbrella:',
    u'\U000026f1': u':umbrella_on_ground:',
    u'\U00002614': u':umbrella_with_rain_drops:',
    u'\U0001f612': u':unamused:',
    u'\U0001f984': u':unicorn_face:',
    u'\U0001f643': u':upside-down_face:',
    u'\U0000270c': u':v:',
    u'\U0001f6a6': u':vertical_traffic_light:',
    u'\U0001f4fc': u':vhs:',
    u'\U0001f4f3': u':vibration_mode:',
    u'\U0001f4f9': u':video_camera:',
    u'\U0001f3ae': u':video_game:',
    u'\U0001f3bb': u':violin:',
    u'\U0000264d': u':virgo:',
    u'\U0001f30b': u':volcano:',
    u'\U0001f3d0': u':volleyball:',
    u'\U0001f318': u':waning_crescent_moon:',
    u'\U0001f316': u':waning_gibbous_moon:',
    u'\U000026a0': u':warning:',
    u'\U0001f5d1': u':wastebasket:',
    u'\U0000231a': u':watch:',
    u'\U0001f403': u':water_buffalo:',
    u'\U0001f6be': u':water_closet:',
    u'\U0001f349': u':watermelon:',
    u'\U0001f44b': u':wave:',
    u'\U0001f3f4': u':waving_black_flag:',
    u'\U0001f3f3': u':waving_white_flag:',
    u'\U00003030': u':wavy_dash:',
    u'\U0001f312': u':waxing_crescent_moon:',
    u'\U0001f629': u':weary:',
    u'\U0001f492': u':wedding:',
    u'\U0001f3cb': u':weight_lifter:',
    u'\U0001f40b': u':whale2:',
    u'\U00002638': u':wheel_of_dharma:',
    u'\U0000267f': u':wheelchair:',
    u'\U00002705': u':white_check_mark:',
    u'\U0001f4ae': u':white_flower:',
    u'\U00002639': u':white_frowning_face:',
    u'\U00002b1c': u':white_large_square:',
    u'\U000025fd': u':white_medium_small_square:',
    u'\U000025fb': u':white_medium_square:',
    u'\U000025ab': u':white_small_square:',
    u'\U0001f533': u':white_square_button:',
    u'\U0001f32c': u':wind_blowing_face:',
    u'\U0001f390': u':wind_chime:',
    u'\U0001f377': u':wine_glass:',
    u'\U0001f609': u':wink:',
    u'\U0001f43a': u':wolf:',
    u'\U0001f469': u':woman:',
    u'\U0001f45a': u':womans_clothes:',
    u'\U0001f452': u':womans_hat:',
    u'\U0001f6ba': u':womens:',
    u'\U0001f5fa': u':world_map:',
    u'\U0001f61f': u':worried:',
    u'\U0001f527': u':wrench:',
    u'\U0000270d': u':writing_hand:',
    u'\U0001f49b': u':yellow_heart:',
    u'\U0000262f': u':yin_yang:',
    u'\U0001f910': u':zipper-mouth_face:',
}

# Expressions taken from Martijn Pieters code at
# http://stackoverflow.com/questions/26568722/remove-unicode-emoji-using-re-in-python
try:
    # Wide UCS-4 build
    ALIAS_RE = re.compile(u'['u'\U0001F300-\U0001F64F'
                          u'\U0001F680-\U0001F6FF'
                          u'\u2600-\u26FF\u2700-\u27BF]+',
                          re.DOTALL | re.UNICODE)
except re.error:
    # Narrow UCS-2 build
    ALIAS_RE = re.compile(u'('u'\ud83c[\udf00-\udfff]|'
                          u'\ud83d[\udc00-\ude4f\ude80-\udeff]|'
                          u'[\u2600-\u26FF\u2700-\u27BF])+',
                          re.DOTALL | re.UNICODE)


def convert_emoji_to_aliases(data, modifier, modifier_data, string):
    if sys.version_info < (3, ):
        string = string.decode('utf-8')
    emoji_found = ALIAS_RE.findall(string)
    for emoji in emoji_found:
        if emoji in EMOJI_ALIASES:
            string = string.replace(emoji, EMOJI_ALIASES[emoji])
    return string


if __name__ == "__main__" and import_ok:

    if w.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE,
                  SCRIPT_DESC, "", "utf-8"):

        w.hook_modifier("irc_in_away", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_cnotice", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_cprivmsg", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_kick", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_knock", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_notice", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_part", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_privmsg", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_quit", "convert_emoji_to_aliases", "")
        w.hook_modifier("irc_in_wallops", "convert_emoji_to_aliases", "")

        # This intercepts outgoing emoji also
        # w.hook_modifier("input_text_for_buffer",
        #                 "convert_emoji_to_aliases", "")



# Script Name: sarcasm.py
# Script Author: Fsaev
# Script License: GPLv3

SCRIPT_NAME = 'sarcasm'
SCRIPT_AUTHOR = 'Fsaev <fredrik@saevland.com>'
SCRIPT_VERSION = '1.0'
SCRIPT_LICENSE = 'GPLv3'
SCRIPT_DESC = 'Adds random capitalization to your sentence'

import_ok = True

try:
    import weechat
except ImportError:
    print('This script must be run under WeeChat')
    print('You can obtain a copy of WeeChat, for free, at https://weechat.org')
    import_ok = False

from random import randint

def sarcasm_cb(data, buffer, args):
    newstring = ""
    for arg in args:
        if randint(0, 1) == 1:
            newstring += arg.upper()
        else:
            newstring += arg.lower()

    weechat.command(buffer, newstring)

    return weechat.WEECHAT_RC_OK

if __name__ == "__main__" and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
        weechat.hook_command(
            "sarcasm",
            """Adds random capitalization to your sentence to indicate that you are being sarcastic, e.g.
/sarcasm I love to put ketchup on my pizza

results in:
i lOVe tO Put KEtChUp oN mY pIzZa
""",
            "message", "",
            "",
            "sarcasm_cb", ""
        )



#
# SPDX-FileCopyrightText: 2009-2026 Sébastien Helleu <flashcode@flashtux.org>
# SPDX-FileCopyrightText: 2010 m4v <lambdae2@gmail.com>
# SPDX-FileCopyrightText: 2011 stfn <stfnmd@googlemail.com>
# SPDX-FileCopyrightText: 2025 Leo Vivier <zaeph@zaeph.net>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Quick jump to buffers."""

import re
from typing import Any

SCRIPT_NAME = "go"
SCRIPT_AUTHOR = "Sébastien Helleu <flashcode@flashtux.org>"
SCRIPT_VERSION = "3.1.1"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "Quick jump to buffers"

SCRIPT_COMMAND = "go"

IMPORT_OK = True

try:
    import weechat  # type: ignore[import]
except ImportError:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: https://weechat.org/")
    IMPORT_OK = False

# script options
SETTINGS = {
    "auto_jump": ("off", "automatically jump to buffer when it is uniquely selected"),
    "buffer_number": ("on", "display buffer number"),
    "color_number": ("yellow,magenta", "color for buffer number (not selected)"),
    "color_number_selected": ("yellow,red", "color for selected buffer number"),
    "color_name": ("black,cyan", "color for buffer name (not selected)"),
    "color_name_selected": ("black,brown", "color for a selected buffer name"),
    "color_name_highlight": (
        "red,cyan",
        "color for highlight in buffer name (not selected)",
    ),
    "color_name_highlight_selected": (
        "red,brown",
        "color for highlight in a selected buffer name",
    ),
    "regexp_search": ("off", "search buffer matches using regexps"),
    "fuzzy_search": ("off", "search buffer matches using approximation"),
    "message": ("Go to: ", "message to display before list of buffers"),
    "min_chars": ("0", "Minimum chars to search and display list of matching buffers"),
    "short_name": ("off", "display and search in short names instead of buffer name"),
    "short_name_server": (
        "off",
        "prefix short names with server names for search and display",
    ),
    "sort": (
        "number,beginning",
        "comma-separated list of keys to sort buffers "
        "(the order is important, sorts are performed in the given order): "
        "name = sort by name (or short name), ",
        "hotlist = sort by hotlist order, "
        "number = first match a buffer number before digits in name, "
        "beginning = first match at beginning of names (or short names); "
        "the default sort of buffers is by numbers",
    ),
    "use_core_instead_weechat": (
        "off",
        'use name "core" instead of "weechat" for core buffer',
    ),
}

# hooks management
HOOK_COMMAND_RUN = {
    "input": ("/input *", "go_command_run_input"),
    "buffer": ("/buffer *", "go_command_run_buffer"),
    "window": ("/window *", "go_command_run_window"),
}
hooks = {}

# input before command /go (we'll restore it later)
saved_input = ""
saved_input_pos = 0

# last user input (if changed, we'll update list of matching buffers)
old_input = None

# matching buffers
buffers = []
buffers_pos = 0


def go_option_enabled(option: str) -> bool:
    """Check if a boolean script option is enabled or not."""
    return weechat.config_string_to_boolean(weechat.config_get_plugin(option))


def go_info_running(_data: str, _info_name: str, _arguments: str) -> str:
    """Return "1" if go is running, otherwise "0"."""
    return "1" if "modifier" in hooks else "0"


def go_unhook_one(hook: str) -> None:
    """Unhook something hooked by this script."""
    global hooks
    if hook in hooks:
        weechat.unhook(hooks[hook])
        del hooks[hook]


def go_unhook_all() -> None:
    """Unhook all."""
    go_unhook_one("modifier")
    for hook in HOOK_COMMAND_RUN:
        go_unhook_one(hook)
    weechat.bar_item_update("input_text")


def go_hook_all() -> None:
    """Hook command_run and modifier."""
    global hooks
    priority = ""
    version = weechat.info_get("version_number", "") or 0
    # use high priority for hook to prevent conflict with other plugins/scripts
    # (WeeChat >= 0.3.4 only)
    if int(version) >= 0x00030400:
        priority = "2000|"
    for hook, value in HOOK_COMMAND_RUN.items():
        if hook not in hooks:
            hooks[hook] = weechat.hook_command_run(
                "%s%s" % (priority, value[0]), value[1], ""
            )
    if "modifier" not in hooks:
        hooks["modifier"] = weechat.hook_modifier(
            "input_text_display_with_cursor", "go_input_modifier", ""
        )
        weechat.bar_item_update("input_text")


def go_start(buf: str) -> None:
    """Start go on buffer."""
    global saved_input, saved_input_pos, old_input, buffers_pos
    go_hook_all()
    saved_input = weechat.buffer_get_string(buf, "input")
    saved_input_pos = weechat.buffer_get_integer(buf, "input_pos")
    weechat.buffer_set(buf, "input", "")
    old_input = None
    buffers_pos = 0


def go_end(buf: str) -> None:
    """End go on buffer."""
    global saved_input, saved_input_pos, old_input
    go_unhook_all()
    weechat.buffer_set(buf, "input", saved_input)
    weechat.buffer_set(buf, "input_pos", str(saved_input_pos))
    old_input = None


def go_match_beginning(buf: str, string: str) -> bool:
    """Check if a string matches the beginning of buffer name/short name."""
    if not string:
        return False
    esc_str = re.escape(string)
    if re.search(r"^#?" + esc_str, buf["name"]) or re.search(
        r"^#?" + esc_str, buf["short_name"]
    ):
        return True
    return False


def go_match_fuzzy(name: str, string: str) -> bool:
    """Check if string matches name using approximation."""
    if not string:
        return False

    name_len = len(name)
    string_len = len(string)

    if string_len > name_len:
        return False
    if name_len == string_len:
        return name == string

    # Attempt to match all chars somewhere in name
    prev_index = -1
    for i, char in enumerate(string):
        index = name.find(char, prev_index + 1)
        if index == -1:
            return False
        prev_index = index
    return True


def go_now(buf: str, args: str) -> None:
    """Go to buffer specified by args."""
    listbuf = go_matching_buffers(args)
    if not listbuf:
        return

    # prefer buffer that matches at beginning (if option is enabled)
    if "beginning" in weechat.config_get_plugin("sort").split(","):
        for index in range(len(listbuf)):
            if go_match_beginning(listbuf[index], args):
                weechat.command(buf, "/buffer " + str(listbuf[index]["full_name"]))
                return

    # jump to first buffer in matching buffers by default
    weechat.command(buf, "/buffer " + str(listbuf[0]["full_name"]))


def go_cmd(_data: str, buf: str, args: str) -> int:
    """Command "/go": just hook what we need."""
    global hooks
    if args:
        go_now(buf, args)
    elif "modifier" in hooks:
        go_end(buf)
    else:
        go_start(buf)
    return weechat.WEECHAT_RC_OK


def go_matching_buffers(strinput: str) -> list[dict[str, Any]]:
    """Return a list with buffers matching user input."""
    global buffers_pos
    listbuf = []
    if len(strinput) == 0:
        buffers_pos = 0
    strinput = strinput.lower()
    infolist = weechat.infolist_get("buffer", "", "")
    while weechat.infolist_next(infolist):
        pointer = weechat.infolist_pointer(infolist, "pointer")
        short_name = weechat.infolist_string(infolist, "short_name")
        server = weechat.buffer_get_string(pointer, "localvar_server")
        if go_option_enabled("short_name"):
            if go_option_enabled("short_name_server") and server:
                name = server + "." + short_name
            else:
                name = short_name
        else:
            name = weechat.infolist_string(infolist, "name")
        if (
            name == "weechat"
            and go_option_enabled("use_core_instead_weechat")
            and weechat.infolist_string(infolist, "plugin_name") == "core"
        ):
            name = "core"
        number = weechat.infolist_integer(infolist, "number")
        full_name = weechat.infolist_string(infolist, "full_name")
        if not full_name:
            full_name = "%s.%s" % (
                weechat.infolist_string(infolist, "plugin_name"),
                weechat.infolist_string(infolist, "name"),
            )
        if go_option_enabled("regexp_search"):
            matching = bool(re.search(strinput, name, re.IGNORECASE))
        else:
            matching = name.lower().find(strinput) >= 0
        if not matching and strinput[-1] == " ":
            matching = name.lower().endswith(strinput.strip())
        if not matching and go_option_enabled("fuzzy_search"):
            matching = go_match_fuzzy(name.lower(), strinput)
        if not matching and strinput.isdigit():
            matching = str(number).startswith(strinput)
        if len(strinput) == 0 or matching:
            listbuf.append(
                {
                    "number": number,
                    "short_name": short_name,
                    "name": name,
                    "full_name": full_name,
                    "pointer": pointer,
                }
            )
    weechat.infolist_free(infolist)

    # sort buffers
    hotlist = []
    infolist = weechat.infolist_get("hotlist", "", "")
    while weechat.infolist_next(infolist):
        hotlist.append(weechat.infolist_pointer(infolist, "buffer_pointer"))
    weechat.infolist_free(infolist)
    last_index_hotlist = len(hotlist)

    def _sort_name(buf):
        """Sort buffers by name (or short name)."""
        return buf["name"]

    def _sort_hotlist(buf):
        """Sort buffers by hotlist order."""
        try:
            return hotlist.index(buf["pointer"])
        except ValueError:
            # not in hotlist, always last.
            return last_index_hotlist

    def _sort_match_number(buf):
        """Sort buffers by match on number."""
        return 0 if str(buf["number"]) == strinput else 1

    def _sort_match_beginning(buf):
        """Sort buffers by match at beginning."""
        return 0 if go_match_beginning(buf, strinput) else 1

    funcs = {
        "name": _sort_name,
        "hotlist": _sort_hotlist,
        "number": _sort_match_number,
        "beginning": _sort_match_beginning,
    }

    for key in weechat.config_get_plugin("sort").split(","):
        if key in funcs:
            listbuf = sorted(listbuf, key=funcs[key])

    if not strinput:
        index = [
            i
            for i, buf in enumerate(listbuf)
            if buf["pointer"] == weechat.current_buffer()
        ]
        if index:
            buffers_pos = index[0]

    return listbuf


def go_buffers_to_string(listbuf: list[dict[str, Any]], pos: int, strinput: str) -> str:
    """Return string built with list of buffers found (matching user input)."""
    try:
        if len(strinput) < int(weechat.config_get_plugin("min_chars")):
            return ""
    except ValueError:
        pass
    string = ""
    strinput = strinput.lower()
    for i in range(len(listbuf)):
        selected = "_selected" if i == pos else ""
        buffer_name = listbuf[i]["name"]
        index = buffer_name.lower().find(strinput)
        if index >= 0:
            index2 = index + len(strinput)
            name = "%s%s%s%s%s" % (
                buffer_name[:index],
                weechat.color(
                    weechat.config_get_plugin("color_name_highlight" + selected)
                ),
                buffer_name[index:index2],
                weechat.color(weechat.config_get_plugin("color_name" + selected)),
                buffer_name[index2:],
            )
        elif go_option_enabled("fuzzy_search") and go_match_fuzzy(
            buffer_name.lower(), strinput
        ):
            name = ""
            prev_index = -1
            for char in strinput.lower():
                index = buffer_name.lower().find(char, prev_index + 1)
                if prev_index < 0:
                    name += buffer_name[:index]
                    name += weechat.color(
                        weechat.config_get_plugin("color_name_highlight" + selected)
                    )
                if prev_index >= 0 and index > prev_index + 1:
                    name += weechat.color(
                        weechat.config_get_plugin("color_name" + selected)
                    )
                    name += buffer_name[prev_index + 1 : index]
                    name += weechat.color(
                        weechat.config_get_plugin("color_name_highlight" + selected)
                    )
                name += buffer_name[index]
                prev_index = index

            name += weechat.color(weechat.config_get_plugin("color_name" + selected))
            name += buffer_name[prev_index + 1 :]
        else:
            name = buffer_name
        string += " "
        if go_option_enabled("buffer_number"):
            string += "%s%s" % (
                weechat.color(weechat.config_get_plugin("color_number" + selected)),
                str(listbuf[i]["number"]),
            )
        string += "%s%s%s" % (
            weechat.color(weechat.config_get_plugin("color_name" + selected)),
            name,
            weechat.color("reset"),
        )
    return "  " + string if string else ""


def go_input_modifier(_data: str, _modifier: str, modifier_data: str, string: str) -> str:
    """This modifier is called when input text item is built by WeeChat.

    This is commonly called after changes in input or cursor move: it builds
    a new input with prefix ("Go to:"), and suffix (list of buffers found).
    """
    global old_input, buffers, buffers_pos
    if modifier_data != weechat.current_buffer():
        return ""
    names = ""
    new_input = weechat.string_remove_color(string, "")
    new_input = new_input.lstrip()
    if old_input is None or new_input != old_input:
        old_buffers = buffers
        buffers = go_matching_buffers(new_input)
        if buffers != old_buffers and len(new_input) > 0:
            if len(buffers) == 1 and go_option_enabled("auto_jump"):
                weechat.command(modifier_data, "/wait 1ms /input return")
            buffers_pos = 0
        old_input = new_input
    names = go_buffers_to_string(buffers, buffers_pos, new_input.strip())
    return weechat.config_get_plugin("message") + string + names


def go_command_run_input(_data: str, buf: str, command: str) -> int:
    """Function called when a command "/input xxx" is run."""
    global buffers, buffers_pos
    if command.startswith("/input search_text") or command.startswith("/input jump"):
        # search text or jump to another buffer is forbidden now
        return weechat.WEECHAT_RC_OK_EAT
    elif command == "/input complete_next":
        # choose next buffer in list
        buffers_pos += 1
        if buffers_pos >= len(buffers):
            buffers_pos = 0
        weechat.hook_signal_send(
            "input_text_changed", weechat.WEECHAT_HOOK_SIGNAL_POINTER, buf
        )
        return weechat.WEECHAT_RC_OK_EAT
    elif command == "/input complete_previous":
        # choose previous buffer in list
        buffers_pos -= 1
        if buffers_pos < 0:
            buffers_pos = len(buffers) - 1
        weechat.hook_signal_send(
            "input_text_changed", weechat.WEECHAT_HOOK_SIGNAL_POINTER, buf
        )
        return weechat.WEECHAT_RC_OK_EAT
    elif command == "/input return":
        # switch to selected buffer (if any)
        go_end(buf)
        if len(buffers) > 0:
            weechat.command(buf, "/buffer " + str(buffers[buffers_pos]["full_name"]))
        return weechat.WEECHAT_RC_OK_EAT
    return weechat.WEECHAT_RC_OK


def go_command_run_buffer(_data: str, _buf: str, _command: str) -> int:
    """Function called when a command "/buffer xxx" is run."""
    return weechat.WEECHAT_RC_OK_EAT


def go_command_run_window(_data: str, buf: str, command: str) -> int:
    """Function called when a command "/window xxx" is run."""
    if command == "/window scroll_bottom":
        # cancel selection and return to input
        go_end(buf)
        return weechat.WEECHAT_RC_OK_EAT
    return weechat.WEECHAT_RC_OK_EAT


def go_unload_script() -> int:
    """Function called when script is unloaded."""
    go_unhook_all()
    return weechat.WEECHAT_RC_OK


def go_main() -> None:
    """Entry point."""
    if not weechat.register(
        SCRIPT_NAME,
        SCRIPT_AUTHOR,
        SCRIPT_VERSION,
        SCRIPT_LICENSE,
        SCRIPT_DESC,
        "go_unload_script",
        "",
    ):
        return
    weechat.hook_command(
        SCRIPT_COMMAND,
        "Quick jump to buffers",
        "[name]",
        "name: directly jump to buffer by name (without argument, list is "
        "displayed)\n\n"
        "You can bind command to a key, for example:\n"
        "  /key bind meta-g /go\n\n"
        "You can use completion key (commonly Tab and shift-Tab) to select "
        "next/previous buffer in list.",
        "%(buffers_names)",
        "go_cmd",
        "",
    )

    # set default settings
    version = weechat.info_get("version_number", "") or 0
    for option, value in SETTINGS.items():
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, value[0])
        if int(version) >= 0x00030500:
            weechat.config_set_desc_plugin(
                option, '%s (default: "%s")' % (value[1], value[0])
            )
    weechat.hook_info(
        "go_running",
        'Return "1" if go is running, otherwise "0"',
        "",
        "go_info_running",
        "",
    )


if __name__ == "__main__" and IMPORT_OK:
    go_main()



# IRCrypt: Secure Encryption Layer Atop IRC
# =========================================
#
# SPDX-FileCopyrightText: 2013-2014 Lars Kiesow <lkiesow@uos.de>
# SPDX-FileCopyrightText: 2013-2014 Sven Haardiek <sven@haardiek.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#
#
# == About ==================================================================
#
#  The weechat IRCrypt plug-in will send messages encrypted to all channels for
#  which a passphrase is set. A channel can either be a regular IRC multi-user
#  channel (i.e. #IRCrypt) or another users nickname.
#
# == Project ================================================================
#
# This plug-in is part of the IRCrypt project. For mor information or to
# participate, please visit
#
#   https://github.com/IRCrypt
#
#
# To report bugs, make suggestions, etc. for this particular plug-in, please
# have a look at:
#
#   https://github.com/IRCrypt/ircrypt-weechat
#

import weechat, string, os, subprocess, base64, time

# Constants used in this script
SCRIPT_NAME    = 'ircrypt'
SCRIPT_AUTHOR  = 'Sven Haardiek <sven@haardiek.de>, Lars Kiesow <lkiesow@uos.de>'
SCRIPT_VERSION = '1.0.1'
SCRIPT_LICENSE = 'GPL3'
SCRIPT_DESC    = 'IRCrypt: Encryption layer for IRC'
SCRIPT_HELP_TEXT = '''
%(bold)sIRCrypt command options: %(normal)s
list                                                    List set keys, public key ids and ciphers
set-key            [-server <server>] <target> <key>    Set key for target
remove-key         [-server <server>] <target>          Remove key for target
set-cipher         [-server <server>] <target> <cipher> Set specific cipher for target
remove-cipher      [-server <server>] <target>          Remove specific cipher for target
plain              [-server <s>] [-channel <ch>] <msg>  Send unencrypted message

%(bold)sExamples: %(normal)s
Set the key for a channel:
   /ircrypt set-key -server freenet #IRCrypt key
Remove the key:
   /ircrypt remove-key #IRCrypt
Switch to a specific cipher for a channel:
   /ircrypt set-cipher -server freenode #IRCrypt TWOFISH
Unset the specific cipher for a channel:
   /ircrypt remove-cipher #IRCrypt
Send unencrypted “Hello” to current channel
   /ircrypt plain Hello

%(bold)sConfiguration: %(normal)s
Tip: You can list all options and what they are currently set to by executing:
   /set ircrypt.*
%(bold)sircrypt.marker.encrypted %(normal)s
   If you add 'ircrypt' to weechat.bar.status.items, these option will set a
   string which is displayed in the status bar of an encrypted channels,
   indicating that the current channel is encrypted.
   If “{{cipher}}” is used as part of this string, it will be replaced by the
   cipher currently used by oneself for that particular channel.
   It is woth noting that you probably don't want to replace the whole value of
   that option but extend it instead in a way like:
      /set weechat.bar.status.items {{currentContent}},ircrypt
%(bold)sircrypt.marker.unencrypted %(normal)s
   This option will set a string which is displayed before each message that is
   send unencrypted in a channel for which a key is set. So you know when
   someone is talking to you without encryption.
%(bold)sircrypt.general.binary %(normal)s
   This will set the GnuPG binary used for encryption and decryption. IRCrypt
   will try to set this automatically.
''' % {'bold':weechat.color('bold'), 'normal':weechat.color('-bold')}

MAX_PART_LEN     = 300
MSG_PART_TIMEOUT = 300 # 5min


# Global variables and memory used to store message parts, pending requests,
# configuration options, keys, etc.
ircrypt_msg_memory       = {}
ircrypt_config_file      = None
ircrypt_config_section   = {}
ircrypt_config_option    = {}
ircrypt_keys             = {}
ircrypt_cipher           = {}
ircrypt_message_plain    = {}


class MessageParts:
    '''Class used for storing parts of messages which were split after
    encryption due to their length.'''

    modified = 0
    last_id  = None
    message  = ''

    def update(self, id, msg):
        '''This method updates an already existing message part by adding a new
        part to the old ones and updating the identifier of the latest received
        message part.
        '''
        # Check if id is correct. If not, throw away old parts:
        if self.last_id and self.last_id != id+1:
            self.message = ''
        # Check if the are old message parts which belong due to their old age
        # probably not to this message:
        if time.time() - self.modified > MSG_PART_TIMEOUT:
            self.message = ''
        self.last_id = id
        self.message = msg + self.message
        self.modified = time.time()


def ircrypt_gnupg(stdin, *args):
    '''Try to execute gpg with given input and options.

    :param stdin: Input for GnuPG
    :param  args: Additional command line options for GnuPG
    :returns:     Tuple containing returncode, stdout and stderr
    '''
    gnupg = weechat.config_string(weechat.config_get('ircrypt.general.binary'))
    if not gnupg:
        return (99, '', 'GnuPG could not be found')
    p = subprocess.Popen(
            [gnupg, '--batch',  '--no-tty'] + list(args),
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate(stdin)
    return (p.returncode, out, err)


def ircrypt_split_msg(cmd, pre, msg):
    '''Convert encrypted message in MAX_PART_LEN sized blocks
    '''
    msg = msg.rstrip()
    return '\n'.join(['%s:>%s-%i %s' %
        (cmd, pre, i // MAX_PART_LEN, msg[i:i+MAX_PART_LEN])
        for i in range(0, len(msg), MAX_PART_LEN)][::-1])


def ircrypt_error(msg, buf):
    '''Print errors to a given buffer. Errors are printed in red and have the
    weechat error prefix.
    '''
    weechat.prnt(buf, weechat.prefix('error') + weechat.color('red') +
            ('\n' + weechat.color('red')).join(msg.split('\n')))


def ircrypt_warn(msg, buf=''):
    '''Print warnings. If no buffer is set, the default weechat buffer is used.
    Warnin are printed in gray without marker.
    '''
    weechat.prnt(buf, weechat.color('gray') +
            ('\n' + weechat.color('gray')).join(msg.split('\n')))


def ircrypt_info(msg, buf=None):
    '''Print ifo message to specified buffer. If no buffer is set, the current
    foreground buffer is used to print the message.
    '''
    if buf is None:
        buf = weechat.current_buffer()
    weechat.prnt(buf, msg)


def ircrypt_decrypt_hook(data, msgtype, server, args):
    '''Hook for incomming PRVMSG commands.
    This method will parse the input, check if it is an encrypted message and
    call the appropriate decryption methods if necessary.

    :param data:
    :param msgtype:
    :param server: IRC server the message comes from.
    :param args: IRC command line-
    '''
    info = weechat.info_get_hashtable('irc_message_parse', { 'message': args })

    # Check if channel is own nick and if change channel to nick of sender
    if info['channel'][0] not in '#&':
        info['channel'] = info['nick']

    # Get key
    key = ircrypt_keys.get(('%s/%s' % (server, info['channel'])).lower())

    # Return everything as it is if we have no key
    if not key:
        return args

    if not '>CRY-' in args:
        # if key exisits and no >CRY not part of message flag message as unencrypted
        pre, message = args.split(' :', 1)
        marker = weechat.config_string(ircrypt_config_option['unencrypted'])
        return '%s :%s %s' % (pre, marker, message)

    # if key exists and >CRY part of message start symmetric encryption
    pre, message    = args.split('>CRY-', 1)
    number, message = message.split(' ', 1 )

    # Get key for the message memory
    catchword = '%s.%s.%s' % (server, info['channel'], info['nick'])

    # Decrypt only if we got last part of the message
    # otherwise put the message into a global memory and quit
    if int(number) != 0:
        if not catchword in ircrypt_msg_memory:
            ircrypt_msg_memory[catchword] = MessageParts()
        ircrypt_msg_memory[catchword].update(int(number), message)
        return ''

    # Get whole message
    try:
        message = message + ircrypt_msg_memory[catchword].message
        del ircrypt_msg_memory[catchword]
    except KeyError:
        pass

    # Get message buffer in case we need to print an error
    buf = weechat.buffer_search('irc', '%s.%s' % (server,info['channel']))

    # Decode base64 encoded message
    try:
        message = base64.b64decode(message)
    except:
        ircrypt_error('Could not Base64 decode message.', buf)
        return args

    # Decrypt
    try:
        message = (key).encode('utf-8') + b'\n' + message
    except:
        # For Python 2.x
        message = key + b'\n' + message
    (ret, out, err) = ircrypt_gnupg(message,
            '--passphrase-fd', '-', '-q', '-d')

    # Get and print GPG errors/warnings
    if ret:
        ircrypt_error(err.decode('utf-8'), buf)
        return args
    if err:
        ircrypt_warn(err.decode('utf-8'))

    return pre + out.decode('utf-8')


def ircrypt_encrypt_hook(data, msgtype, server, args):
    '''Hook for outgoing PRVMSG commands.
    This method will call the appropriate methods for encrypting the outgoing
    messages either symmetric or asymmetric

    :param data:
    :param msgtype:
    :param server: IRC server the message comes from.
    :param args: IRC command line-
    '''
    info = weechat.info_get_hashtable("irc_message_parse", { "message": args })

    # check if this message is to be send as plain text
    plain = ircrypt_message_plain.get('%s/%s' % (server, info['channel']))
    if plain:
        del ircrypt_message_plain['%s/%s' % (server, info['channel'])]
        if (plain[0] - time.time()) < 5 \
                and args == 'PRIVMSG %s :%s' % (info['channel'], plain[1]):
            args = args.replace('PRIVMSG %s :%s ' % (
                info['channel'],
                weechat.config_string(ircrypt_config_option['unencrypted'])),
                'PRIVMSG %s :' % info['channel'])
            return args

    # check symmetric key
    key = ircrypt_keys.get(('%s/%s' % (server, info['channel'])).lower())
    if not key:
        # No key -> don't encrypt
        return args

    # Get cipher
    cipher = ircrypt_cipher.get(('%s/%s' % (server, info['channel'])).lower(),
            weechat.config_string(ircrypt_config_option['sym_cipher']))
    # Get prefix and message
    pre, message = args.split(':', 1)

    # encrypt message
    try:
        inp = key.encode('utf-8') + b'\n' + message.encode('utf-8')
    except:
        inp = key + b'\n' + message
    (ret, out, err) = ircrypt_gnupg(inp,
            '--symmetric', '--cipher-algo', cipher, '--passphrase-fd', '-')

    # Get and print GPG errors/warnings
    if ret:
        buf = weechat.buffer_search('irc', '%s.%s' % (server, info['channel']))
        ircrypt_error(err.decode('utf-8'), buf)
        return args
    if err:
        ircrypt_warn(err.decode('utf-8'))

    # Ensure the generated messages are not too long and send them
    return ircrypt_split_msg(pre, 'CRY', base64.b64encode(out).decode('utf-8'))


def ircrypt_config_init():
    ''' This method initializes the configuration file. It creates sections and
    options in memory and prepares the handling of key sections.
    '''
    global ircrypt_config_file
    ircrypt_config_file = weechat.config_new('ircrypt', 'ircrypt_config_reload_cb', '')
    if not ircrypt_config_file:
        return

    # marker
    ircrypt_config_section['marker'] = weechat.config_new_section(
            ircrypt_config_file, 'marker', 0, 0, '', '', '', '', '', '', '', '',
            '', '')
    if not ircrypt_config_section['marker']:
        weechat.config_free(ircrypt_config_file)
        return
    ircrypt_config_option['encrypted'] = weechat.config_new_option(
            ircrypt_config_file, ircrypt_config_section['marker'],
            'encrypted', 'string', 'Marker for encrypted messages', '', 0, 0,
            'encrypted', 'encrypted', 0, '', '', '', '', '', '')
    ircrypt_config_option['unencrypted'] = weechat.config_new_option(
            ircrypt_config_file, ircrypt_config_section['marker'], 'unencrypted',
            'string', 'Marker for unencrypted messages received in an encrypted channel',
            '', 0, 0, '', 'u', 0, '', '', '', '', '', '')

    # cipher options
    ircrypt_config_section['cipher'] = weechat.config_new_section(
            ircrypt_config_file, 'cipher', 0, 0, '', '', '', '', '', '', '', '',
            '', '')
    if not ircrypt_config_section['cipher']:
        weechat.config_free(ircrypt_config_file)
        return
    ircrypt_config_option['sym_cipher'] = weechat.config_new_option(
            ircrypt_config_file, ircrypt_config_section['cipher'],
            'sym_cipher', 'string', 'symmetric cipher used by default', '', 0, 0,
            'TWOFISH', 'TWOFISH', 0, '', '', '', '', '', '')

    # general options
    ircrypt_config_section['general'] = weechat.config_new_section(
            ircrypt_config_file, 'general', 0, 0, '', '', '', '', '', '', '', '',
            '', '')
    if not ircrypt_config_section['general']:
        weechat.config_free(ircrypt_config_file)
        return
    ircrypt_config_option['binary'] = weechat.config_new_option(
            ircrypt_config_file, ircrypt_config_section['general'],
            'binary', 'string', 'GnuPG binary to use', '', 0, 0,
            '', '', 0, '', '', '', '', '', '')

    # keys
    ircrypt_config_section['keys'] = weechat.config_new_section(
            ircrypt_config_file, 'keys', 0, 0, 'ircrypt_config_keys_read_cb', '',
            'ircrypt_config_keys_write_cb', '', '', '', '', '', '', '')
    if not ircrypt_config_section['keys']:
        weechat.config_free(ircrypt_config_file)

    # Special Ciphers
    ircrypt_config_section['special_cipher'] = weechat.config_new_section(
            ircrypt_config_file, 'special_cipher', 0, 0,
            'ircrypt_config_special_cipher_read_cb', '',
            'ircrypt_config_special_cipher_write_cb', '', '', '', '', '', '', '')
    if not ircrypt_config_section['special_cipher']:
        weechat.config_free(ircrypt_config_file)


def ircrypt_config_reload_cb(data, config_file):
    '''Handle a reload of the configuration file.
    '''
    global ircrypt_keys, ircrypt_cipher
    # Forget Keys and ciphers to make sure they are properly reloaded and no old
    # ones are left
    ircrypt_keys   = {}
    ircrypt_cipher = {}
    return weechat.config_reload(config_file)


def ircrypt_config_read():
    ''' Read IRCrypt configuration file (ircrypt.conf).
    '''
    return weechat.config_read(ircrypt_config_file)


def ircrypt_config_write():
    ''' Write IRCrypt configuration file (ircrypt.conf) to disk.
    '''
    return weechat.config_write(ircrypt_config_file)


def ircrypt_config_keys_read_cb(data, config_file, section_name, option_name,
        value):
    '''Read elements of the key section from the configuration file.
    '''
    ircrypt_keys[option_name.lower()] = value
    return weechat.WEECHAT_CONFIG_OPTION_SET_OK_CHANGED


def ircrypt_config_keys_write_cb(data, config_file, section_name):
    '''Write passphrases to the key section of the configuration file.
    '''
    weechat.config_write_line(config_file, section_name, '')
    for target, key in sorted(list(ircrypt_keys.items())):
        weechat.config_write_line(config_file, target.lower(), key)

    return weechat.WEECHAT_RC_OK


def ircrypt_config_special_cipher_read_cb(data, config_file, section_name,
        option_name, value):
    '''Read elements of the key section from the configuration file.
    '''
    ircrypt_cipher[option_name.lower()] = value
    return weechat.WEECHAT_CONFIG_OPTION_SET_OK_CHANGED


def ircrypt_config_special_cipher_write_cb(data, config_file, section_name):
    '''Write passphrases to the key section of the configuration file.
    '''
    weechat.config_write_line(config_file, section_name, '')
    for target, cipher in sorted(list(ircrypt_cipher.items())):
        weechat.config_write_line(config_file, target.lower(), cipher)
    return weechat.WEECHAT_RC_OK


def ircrypt_command_list():
    '''List set keys and channel specific ciphers.
    '''
    # List keys
    keys = '\n'.join([' %s : %s' % x for x in ircrypt_keys.items()])
    ircrypt_info('Symmetric Keys:\n' + keys if keys else 'No symmetric keys set')

    # List channel specific ciphers
    ciphers = '\n'.join([' %s : %s' % x for x in ircrypt_cipher.items()])
    ircrypt_info('Special ciphers:\n' + ciphers if ciphers
            else 'No special ciphers set')
    return weechat.WEECHAT_RC_OK


def ircrypt_command_set_keys(target, key):
    '''Set key for target.

    :param target: server/channel combination
    :param key: Key to use for target
    '''
    ircrypt_keys[target.lower()] = key
    ircrypt_info('Set key for %s' % target)
    return weechat.WEECHAT_RC_OK


def ircrypt_command_remove_keys(target):
    '''Remove key for target.

    :param target: server/channel combination
    '''
    try:
        del ircrypt_keys[target.lower()]
        ircrypt_info('Removed key for %s' % target)
    except KeyError:
        ircrypt_info('No existing key for %s.' % target)
    return weechat.WEECHAT_RC_OK


def ircrypt_command_set_cip(target, cipher):
    '''Set cipher for target.

    :param target: server/channel combination
    :param cipher: Cipher to use for target
    '''
    ircrypt_cipher[target.lower()] = cipher
    ircrypt_info('Set cipher %s for %s' % (cipher, target))
    return weechat.WEECHAT_RC_OK


def ircrypt_command_remove_cip(target):
    '''Remove cipher for target.

    :param target: server/channel combination
    '''
    try:
        del ircrypt_cipher[target.lower()]
        ircrypt_info('Removed special cipher. Using default cipher for %s instead.' % target)
    except KeyError:
        ircrypt_info('No special cipher set for %s.' % target)
    return weechat.WEECHAT_RC_OK


def ircrypt_command_plain(buffer, server, args, argv):
    '''Send unencrypted message
    '''
    channel = ''
    if (len(argv) > 2 and argv[1] == '-channel'):
        channel = argv[2]
        args = (args.split(' ', 2)+[''])[2]
    else:
        # Try to determine the server automatically
        channel = weechat.buffer_get_string(buffer, 'localvar_channel')
    # If there is no text, just ignore the command
    if not args:
        return weechat.WEECHAT_RC_OK
    marker = weechat.config_string(ircrypt_config_option['unencrypted'])
    msg = marker + ' ' + args.split(' ', 1)[-1]
    ircrypt_message_plain['%s/%s' % (server, channel)] = (time.time(), msg)
    weechat.command('','/msg -server %s %s %s' % \
            (server, channel, msg))
    return weechat.WEECHAT_RC_OK


def ircrypt_command(data, buffer, args):
    '''Hook to handle the /ircrypt weechat command.
    '''
    argv = args.split()

    # list
    if not argv or argv == ['list']:
        return ircrypt_command_list()

    # Check if a server was set
    if (len(argv) > 2 and argv[1] == '-server'):
        server = argv[2]
        del argv[1:3]
        args = (args.split(' ', 2)+[''])[2]
    else:
        # Try to determine the server automatically
        server = weechat.buffer_get_string(buffer, 'localvar_server')

    # All remaining commands need a server name
    if not server:
        ircrypt_error('Unknown Server. Please use -server to specify server', buffer)
        return weechat.WEECHAT_RC_ERROR

    if argv[:1] == ['plain']:
        return ircrypt_command_plain(buffer, server, args, argv)

    try:
        target = '%s/%s' % (server, argv[1])
    except:
        ircrypt_error('Unknown command. Try  /help ircrypt', buffer)
        return weechat.WEECHAT_RC_OK

    # Set keys
    if argv[:1] == ['set-key']:
        if len(argv) < 3:
            return weechat.WEECHAT_RC_ERROR
        return ircrypt_command_set_keys(target, ' '.join(argv[2:]))

    # Remove keys
    if argv[:1] == ['remove-key']:
        if len(argv) != 2:
            return weechat.WEECHAT_RC_ERROR
        return ircrypt_command_remove_keys(target)

    # Set special cipher for channel
    if argv[:1] == ['set-cipher']:
        if len(argv) < 3:
            return weechat.WEECHAT_RC_ERROR
        return ircrypt_command_set_cip(target, ' '.join(argv[2:]))

    # Remove secial cipher for channel
    if argv[:1] == ['remove-cipher']:
        if len(argv) != 2:
            return weechat.WEECHAT_RC_ERROR
        return ircrypt_command_remove_cip(target)

    ircrypt_error('Unknown command. Try  /help ircrypt', buffer)
    return weechat.WEECHAT_RC_OK


def ircrypt_encryption_statusbar(*args):
    '''This method will set the “ircrypt” element of the status bar if
    encryption is enabled for the current channel. The placeholder {{cipher}}
    can be used, which will be replaced with the cipher used for the current
    channel.
    '''
    channel = weechat.buffer_get_string(weechat.current_buffer(), 'localvar_channel')
    server  = weechat.buffer_get_string(weechat.current_buffer(), 'localvar_server')
    key = ircrypt_keys.get(('%s/%s' % (server, channel)).lower())

    # Return nothing if no key is set for current channel
    if not key:
        return ''

    # Get cipher used for current channel
    cipher = weechat.config_string(ircrypt_config_option['sym_cipher'])
    cipher = ircrypt_cipher.get(('%s/%s' % (server, channel)).lower(), cipher)

    # Return marker, but replace {{cipher}}
    marker = weechat.config_string(ircrypt_config_option['encrypted'])
    return marker.replace('{{cipher}}', cipher)


def ircrypt_find_gpg_binary(names=('gpg2','gpg')):
    '''Check for GnuPG binary to use
    :returns: Tuple with binary name and version.
    '''
    for binary in names:
        p = subprocess.Popen([binary, '--version'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        version = p.communicate()[0].decode('utf-8').split('\n',1)[0]
        if not p.returncode:
            return binary, version
    return None, None


def ircrypt_check_binary():
    '''If binary is not set, try to determine it automatically
    '''
    cfg_option = weechat.config_get('ircrypt.general.binary')
    gnupg = weechat.config_string(cfg_option)
    if not gnupg:
        (gnupg, version) = ircrypt_find_gpg_binary(('gpg','gpg2'))
        if not gnupg:
            ircrypt_error('Automatic detection of the GnuPG binary failed and '
                    'nothing is set manually. You wont be able to use IRCrypt like '
                    'this. Please install GnuPG or set the path to the binary to '
                    'use.', '')
        else:
            ircrypt_info('Found %s' % version, '')
            weechat.config_option_set(cfg_option, gnupg, 1)


# register plugin
if __name__ == '__main__' and weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR,
        SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, 'ircrypt_unload_script',
        'UTF-8'):
    # register the modifiers
    ircrypt_config_init()
    ircrypt_config_read()
    ircrypt_check_binary()
    weechat.hook_modifier('irc_in_privmsg',  'ircrypt_decrypt_hook', '')
    weechat.hook_modifier('irc_out_privmsg', 'ircrypt_encrypt_hook', '')

    weechat.hook_command('ircrypt', 'Commands to manage IRCrypt options and execute IRCrypt commands',
            '[list]'
            '| set-key [-server <server>] <target> <key> '
            '| remove-key [-server <server>] <target> '
            '| set-cipher [-server <server>] <target> <cipher> '
            '| remove-cipher [-server <server>] <target> '
            '| plain [-server <server>] [-channel <channel>] <message>',
            SCRIPT_HELP_TEXT,
            'list || set-key %(irc_channel)|%(nicks)|-server %(irc_servers) %- '
            '|| remove-key %(irc_channel)|%(nicks)|-server %(irc_servers) %- '
            '|| set-cipher %(irc_channel)|-server %(irc_servers) %- '
            '|| remove-cipher |%(irc_channel)|-server %(irc_servers) %- '
            '|| plain |-channel %(irc_channel)|-server %(irc_servers) %-',
            'ircrypt_command', '')
    weechat.bar_item_new('ircrypt', 'ircrypt_encryption_statusbar', '')
    weechat.hook_signal('ircrypt_buffer_opened', 'update_encryption_status', '')


def ircrypt_unload_script():
    '''Hook to ensure the configuration is properly written to disk when the
    script is unloaded.
    '''
    ircrypt_config_write()
    return weechat.WEECHAT_RC_OK



# -*- coding: utf-8 -*-
# Author: Ryan Feng <odayfans at gmail dot com>
# License: GPL3
# Version: 0.2
# Changelog
# 0.2:
# * Fewer error messages

import gntp.notifier as notifier
import weechat
import time

# Logging
def log(msg):
    weechat.prnt("",msg)

# Register plugin
weechat.register("gntpnotify",
             "Ryan Feng",
             "0.2",
             "GPL3",
             "GNTP Notify: Growl notifications using python-gntp",
             "",
             "")

# Create hooks
weechat.hook_signal("*,irc_in2_PRIVMSG","private_msg","")
weechat.hook_signal("weechat_highlight","hilight_msg","")

growl = None
grow_loaded = False
last_log_time = None
log_period = 10 # Redisplay error log in 10s
error_log_max = 3  # Only show no more than 3 error messages in 10s
error_logged = 0

def connect():
    global growl, grow_loaded, log_period, error_log_max, last_log_time, error_logged
    # Create grow object
    growl = notifier.GrowlNotifier(
        applicationName = "Weechat",
        notifications = ["irc message"],
        defaultNotifications = ["irc message"],
        )
    try:
        grow_loaded = growl.register()
    except:
        grow_loaded = False
        if last_log_time is None:
            last_log_time = time.time()
        elapsed_time = time.time() - last_log_time
        if elapsed_time >= log_period:
            error_logged = 0

        if error_logged < error_log_max:
            log("Cannot create notifier object, please make sure Growl has started")
            error_logged += 1
            last_log_time = time.time()

    return grow_loaded

def show_notification(title, message):
    if not grow_loaded:
        connect()
        return
    r = growl.notify(
            noteType = "irc message",
            title = title,
            description = message,
            sticky = False,
            priority = 1
            )
    if not r:
        log("Cannot send notification")

def private_msg(data, signal, message):
    # message is a raw irc message sent from the server
    message = message[1:]

    # Get sender
    sender = message[:message.find('!')-1]

    # Get receiver
    receiver = message.split()[2]

    # Get message body
    msg = message[message.find(':')+1:]

    # Get the server which sent this message
    server = signal.partition(',')[0]

    # Ignore all PRIVMSGs send to a channel,the rest are messages send to 'me'
    if not receiver.startswith('#'):
        show_notification("%s @ #%s" % (sender,server),msg)

    return weechat.WEECHAT_RC_OK

def hilight_msg(data, signal, message):
    # TODO: Get the sender of the message and the channel where the highlight displayed
    sender,_,tmp_msg = message.partition('\t')

    _,_,msg = tmp_msg.partition(' ')

    if grow_loaded:
        show_notification("%s" % sender, msg)
    else:
        connect()
    return weechat.WEECHAT_RC_OK




import weechat, re

SCRIPT_NAME    = "glitter"
SCRIPT_AUTHOR  = "jotham.read@gmail.com"
SCRIPT_VERSION = "0.1"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Replaces ***text*** you write with rainbow text"

if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
   weechat.hook_command_run("/input return", "command_run_input", "")

glitter_pat = re.compile("\*\*\*([^\*]+)\*\*\*")
def glitter_it(match):
   lut = ("13","4","8","9","11","12") # len=6
   text = match.group(1)
   return "".join(["\03"+lut[i%6]+text[i] for i in range(len(text))]) + "\03"

def command_run_input(data, buffer, command):
   if command == "/input return":
      input = weechat.buffer_get_string(buffer, 'input')
      if input.startswith('/set '): # Skip modification of settings
         return weechat.WEECHAT_RC_OK
      input = glitter_pat.sub(glitter_it, input)
      weechat.buffer_set(buffer, 'input', input)
   return weechat.WEECHAT_RC_OK



#
# Copyright (c) 2009 - 2018 by kinabalu (https://mysticcoders.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Kickban script with a timeout, waits for N seconds and unbans user
#
# History:
#
# 2022-01-25, Sébastien Helleu <flashcode@flashtux.org>
#       version 0.3, fix mixed spaces and tabs for indentation
# 2018-07-22, kinabalu (https://mysticcoders.com)
#       version 0.2, script cleanup and ensure py3 compatibility
# 2009-05-03, kinabalu
#       version 0.1, initial version
#
# /kbtimeout and /kbt can be used in combination with these actions
#
# /kbt [nick] [timeout] [message]
# /help kbt
#
# The script can be laoded into WeeChat by executing:
#
# /python load kbtimeout.py
#
# The script may also be auto-loaded by WeeChat.  See the
# WeeChat manual for instructions about how to do this.
#
# For up-to-date information about this script, and new
# version downloads, please go to:
#
# https://www.mysticcoders.com/
#
# If you have any questions, please contact me on-line at:
#
# irc.freenode.net - kinabalu (op): ##java
#
# - kinabalu
#

import os
import string

try:
    import weechat
except ImportError:
    print('This script has to run under WeeChat (https://weechat.org/).')
    sys.exit(1)

SCRIPT_NAME = 'kbtimeout'
SCRIPT_AUTHOR = 'kinabalu (https://mysticcoders.com)'
SCRIPT_VERSION = '0.3'
SCRIPT_LICENSE = 'GPL3'

def handler(data, buffer, argList):

  split_args = argList.split(" ")
  if len(split_args) < 2:
    weechat.prnt("", "Wrong number of parameters for kbtimeout")
    return weechat.WEECHAT_RC_ERROR;

  nick = split_args[0]
  timeout = split_args[1]

  message = ""
  if len(split_args) > 2:
    message = argList.split(" ", 2)[2]

  nick_ptr = weechat.nicklist_search_nick(buffer, "", nick)
  infolist = weechat.infolist_get("irc_nick", "", "{},{}".format("freenode", "##kbtimeout"))

  buffer_name = weechat.buffer_get_string(buffer, "name");

  found_ban_host = None
  while weechat.infolist_next(infolist):
      ban_nick = weechat.infolist_string(infolist, "name")
      ban_host = weechat.infolist_string(infolist, "host")
      ban_account = weechat.infolist_string(infolist, "account")

      if ban_nick == nick:
          found_ban_host = ban_host
  weechat.infolist_free(infolist)

  found_ban_host = found_ban_host[1:]

  if found_ban_host:
    weechat.command(buffer, "/kickban " + found_ban_host + " " + message)
    weechat.hook_timer(int(timeout) * 1000, 0, 1, "kickban_callback", buffer_name + ":" + found_ban_host)

  return weechat.WEECHAT_RC_OK
# END handler

def kickban_callback(data, times_left):
  details = data.split(":")
  buffer_ptr = weechat.buffer_search("irc", details[0])

  if buffer_ptr:
    weechat.command(buffer_ptr, "/unban " + details[1])
  return weechat.WEECHAT_RC_OK
# END kickban_callback

if __name__ == '__main__':
    weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, "kickban with timeout", "", "")

    weechat.hook_command(SCRIPT_NAME, "Kickban with ban timeout", "[nick] [timeout] [comment]", "kickban nick with comment", "%(irc_channel_nicks_hosts) %-", "handler", "")
    weechat.hook_command("kbt", "Kickban with ban timeout", "[nick] [timeout] [comment]", "kickban nick with comment", "%(irc_channel_nicks_hosts) %-", "handler", "")



#
# SPDX-FileCopyrightText: 2012-2026 Sébastien Helleu <flashcode@flashtux.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Interactive maze generator and solver for WeeChat."""

import random
from dataclasses import dataclass, field

try:
    import weechat  # type: ignore[import]

    IMPORT_OK = True
except ImportError:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: https://weechat.org/")
    IMPORT_OK = False

SCRIPT_NAME = "maze"
SCRIPT_AUTHOR = "Sébastien Helleu <flashcode@flashtux.org>"
SCRIPT_VERSION = "1.0.1"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "Interactive maze generator and solver"

SCRIPT_COMMAND = "maze"

MAZE_KEYS: dict[str, tuple[str, str]] = {
    "n": ("", "new maze"),
    "d": ("default", "default size"),
    "+": ("+", "bigger"),
    "-": ("-", "smaller"),
    "s": ("solve", "solve"),
    "i": ("isolve", "solve interactively"),
    "r": ("reset", "reset solution"),
}


@dataclass
class Maze:
    """Maze."""

    width: int = 0
    height: int = 0

    cells: list[int] = field(default_factory=list, init=False)
    solution: list[tuple[int, int, bool]] = field(default_factory=list, init=False)
    buffer: str = field(init=False)
    timer: str = field(default="", init=False)

    # cell status
    VISITED: int = 1  # visited cell
    TOP: int = 2  # door opened on top
    BOTTOM: int = 4  # door opened on bottom
    LEFT: int = 8  # door opened on the left
    RIGHT: int = 16  # door opened on the right
    SOLUTION: int = 32  # cell part of the solution
    SOLUTION_INT: int = 64  # cell displayed for solution

    def __post_init__(self) -> None:
        """Initialize maze with the given size."""
        self.cells = [0] * (self.width * self.height)
        self.cells[0] |= self.LEFT
        self.cells[-1] |= self.RIGHT
        self.solution = []
        self.buffer = Maze.open_buffer()
        if self.buffer:
            keys = ", ".join(
                [f'alt+"{key}": {value[1]}' for key, value in MAZE_KEYS.items()]
            )
            weechat.buffer_set(
                self.buffer,
                "title",
                f"Maze {self.width}x{self.height} | {keys}",
            )

    @staticmethod
    def open_buffer() -> str:
        """Open maze buffer."""
        buf: str = weechat.buffer_search("python", "maze")
        if not buf:
            buf = weechat.buffer_new(
                "maze",
                "maze_input_buffer",
                "",
                "maze_close_buffer",
                "",
            )
        if buf:
            weechat.buffer_set(buf, "type", "free")
            for key, value in MAZE_KEYS.items():
                weechat.buffer_set(
                    buf,
                    f"key_bind_meta-{key}",
                    f"/{SCRIPT_COMMAND} {value[0]}".strip(),
                )
            weechat.buffer_set(buf, "display", "1")
        return buf

    def get_adjacent_cells(
        self, col: int, line: int
    ) -> list[tuple[int, int, int, int]]:
        """Get adjacent cells."""
        list_cells = []
        if col < self.width - 1:  # right
            list_cells.append((col + 1, line, self.RIGHT, self.LEFT))
        if line < self.height - 1:  # bottom
            list_cells.append((col, line + 1, self.BOTTOM, self.TOP))
        if line > 0:  # top
            list_cells.append((col, line - 1, self.TOP, self.BOTTOM))
        if col > 0:  # left
            list_cells.append((col - 1, line, self.LEFT, self.RIGHT))
        return list_cells

    def generate(self) -> None:
        """Generate a new random maze."""
        if not self.cells:
            return
        col: int = 0
        line: int = 0
        self.cells[0] |= self.VISITED
        stack: list[tuple[int, int]] = []
        while True:
            pos: int = (line * self.width) + col
            list_cells: list[tuple[int, int, int, int]] = self.get_adjacent_cells(
                col, line
            )
            random.shuffle(list_cells)
            for col2, line2, to_neighbor, from_neighbor in list_cells:
                pos2 = (line2 * self.width) + col2
                # neighbor not visited?
                if not self.cells[pos2] & self.VISITED:
                    # open door from (x, y) to neighbor
                    self.cells[pos] |= to_neighbor
                    # open door from neighbor to (x, y) (reverse)
                    self.cells[pos2] |= from_neighbor | self.VISITED
                    stack.append((col, line))
                    col, line = col2, line2
                    break
            else:
                col, line = stack.pop()
                if not stack:
                    break

    def solve(self, interactive: bool = False) -> None:
        """Solve a maze: find path from entry (0,0) to exit (n,n)."""
        self.remove_timer()
        for index, cell in enumerate(self.cells):
            self.cells[index] = cell & ~(
                self.VISITED | self.SOLUTION | self.SOLUTION_INT
            )
        col: int = 0
        line: int = 0
        self.cells[0] |= self.SOLUTION
        self.solution = [(col, line, True)]
        stack: list[tuple[int, int]] = []
        visited_solution = self.VISITED | self.SOLUTION
        while not self.cells[-1] & self.SOLUTION:
            pos: int = (line * self.width) + col
            self.cells[pos] |= self.VISITED
            list_cells: list[tuple[int, int, int, int]] = self.get_adjacent_cells(
                col, line
            )
            for col2, line2, to_neighbor, _ in list_cells:
                pos2: int = (line2 * self.width) + col2
                # door opened and neighbor not visited/solution?
                if (
                    self.cells[pos] & to_neighbor
                    and not self.cells[pos2] & visited_solution
                ):
                    self.cells[pos2] |= visited_solution
                    self.solution.append((col2, line2, True))
                    stack.append((col, line))
                    col, line = col2, line2
                    break
            else:
                if not self.cells[-1] & self.SOLUTION:
                    self.cells[pos] &= ~(self.SOLUTION)
                    self.solution.append((col, line, False))
                col, line = stack.pop()
                if not stack:
                    break
        self.display()
        if interactive:
            self.show_interactive_solution()

    def reset(self) -> None:
        """Remove solution."""
        self.remove_timer()
        for index, cell in enumerate(self.cells):
            self.cells[index] = cell & ~(self.SOLUTION | self.SOLUTION_INT)
        self.display()

    def display_line(self, line: int) -> None:
        """Display a line of maze."""
        str_line: str
        if line >= self.height:
            str_line = weechat.color("white") + ("▀" * ((self.width * 2) + 1))
        else:
            both_sol: int = self.SOLUTION | self.SOLUTION_INT
            cell: int = self.cells[line * self.width]
            color: str = (
                "lightred"
                if cell & self.SOLUTION_INT
                else "blue"
                if cell & self.SOLUTION
                else "black"
            )
            str_line = weechat.color(f"{color},white") + (
                "▄" if cell & self.LEFT else " "
            )
            for col in range(self.width):
                cell = self.cells[(line * self.width) + col]
                color = (
                    "lightred"
                    if cell & self.SOLUTION_INT
                    else "blue"
                    if cell & self.SOLUTION
                    else "black"
                )
                # door opened on top?
                if cell & self.TOP:
                    pos: int = ((line - 1) * self.width) + col
                    if line > 0 and self.cells[pos] & both_sol:
                        str_line += weechat.color(f"white,{color}") + " "
                    else:
                        str_line += weechat.color(f"{color},black") + "▄"
                else:
                    str_line += weechat.color(f"{color},white") + "▄"
                # door opened on the right?
                str_line += weechat.color(f"{color},white") + (
                    "▄" if cell & self.RIGHT else " "
                )
        str_line += weechat.color("reset")
        weechat.prnt_y(self.buffer, line, str_line)

    def display(self) -> None:
        """Display maze."""
        if not self.buffer:
            return
        weechat.buffer_clear(self.buffer)
        for line in range(self.height + 1):
            self.display_line(line)

    def remove_timer(self) -> None:
        """Reset the timer used to show interactive solution."""
        if self.timer:
            weechat.unhook(self.timer)
            self.timer = ""

    def show_interactive_solution(self) -> None:
        """Show solution."""
        self.remove_timer()
        self.timer = weechat.hook_timer(1, 0, 0, "maze_timer_cb", "")

    def show_next_solution(self) -> None:
        """Show next solution step."""
        col, line, show = self.solution.pop(0)
        pos: int = (line * self.width) + col
        if show:
            self.cells[pos] |= self.SOLUTION_INT
        else:
            self.cells[pos] &= ~(self.SOLUTION_INT)
        self.display_line(line)
        if not self.solution:
            self.remove_timer()

    def __del__(self) -> None:
        """Destructor."""
        self.remove_timer()


maze: Maze | None = None


def maze_input_buffer(_data: str, buffer: str, str_input: str) -> int:
    """Input data in maze buffer."""
    # pylint: disable=unused-argument
    if str_input.lower() == "q":
        weechat.buffer_close(buffer)
    else:
        weechat.command("", f"/{SCRIPT_COMMAND} {str_input}")
    return weechat.WEECHAT_RC_OK


def maze_close_buffer(_data: str, _buffer: str) -> int:
    """Called when maze buffer is closed."""
    # pylint: disable=unused-argument
    global maze
    maze = None
    return weechat.WEECHAT_RC_OK


def maze_timer_cb(_data: str, _remaining_calls: int) -> int:
    """Timer used to show solution, one cell by one cell."""
    global maze
    if maze:
        maze.show_next_solution()
    return weechat.WEECHAT_RC_OK


def maze_get_size(args: str = "") -> tuple[int, int]:
    """Get maze size with args, defaulting to current maze or window size."""
    global maze
    width: int = 0
    height: int = 0
    if args in ("d", "default"):
        args = ""
    elif maze:
        width, height = maze.width, maze.height
    if args:
        # size given by user
        try:
            items = args.split()
            if len(items) > 1:
                width, height = int(items[0]), int(items[1])
            else:
                width, height = int(args), int(args)
            width = max(width, 2)
            height = max(height, 2)
        except ValueError:
            width, height = 0, 0
    if not width or not height:
        # automatic size with size of window
        win_width: int = (
            weechat.window_get_integer(weechat.current_window(), "win_chat_width") - 1
        )
        win_height: int = (
            weechat.window_get_integer(weechat.current_window(), "win_chat_height") - 1
        )
        size: int = min(win_width, win_height)
        width, height = size, size
    return width, height


def maze_get_other_size(pct_diff: int) -> tuple[int, int]:
    """Get another size using a percent of size to add or subtract."""
    factor: int = pct_diff // abs(pct_diff)
    width, height = maze_get_size()
    add_width: int = max(2, (width * abs(pct_diff)) // 100)
    add_height: int = max(2, (height * abs(pct_diff)) // 100)
    width += factor * add_width
    height += factor * add_height
    return max(width, 2), max(height, 2)


def maze_new(width: int, height: int) -> None:
    """Create a new maze."""
    global maze
    maze = Maze(width=width, height=height)
    maze.generate()
    maze.display()


def maze_cmd_cb(_data: str, _buffer: str, args: str) -> int:
    """The /maze command."""
    global maze
    if args in ("s", "solve"):
        if maze:
            maze.solve()
    elif args in ("i", "isolve"):
        if maze:
            maze.solve(interactive=True)
    elif args in ("r", "reset"):
        if maze:
            maze.reset()
    else:
        if args == "+":
            width, height = maze_get_other_size(30)
        elif args == "-":
            width, height = maze_get_other_size(-30)
        elif args in ("d", "default"):
            width, height = maze_get_size(args)
        elif not args or args in ("n", "new"):
            width, height = maze_get_size()
        else:
            error = weechat.prefix("error")
            weechat.prnt("", f'{error}maze error: unknown option "{args}"')
            return weechat.WEECHAT_RC_OK
        maze_new(width, height)
    return weechat.WEECHAT_RC_OK


if (
    __name__ == "__main__"
    and IMPORT_OK
    and weechat.register(
        SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""
    )
):
    weechat.hook_command(
        SCRIPT_COMMAND,
        "Generate and solve a maze.",
        "size || n|new || d|default || s|solve || i|isolve || r|reset || +|-",
        """\
   size: one size (square) or width and height separated by spaces
    new: regenerate another maze
default: regenerate another maze with default size
  solve: solve the maze and show solution (in blue)
 isolve: solve then show interactively how solution was found (in red)
  reset: hide solution
      +: show a bigger maze
      -: show a smaller maze

All options shown above can be given in input of maze buffer.
""",
        "",
        "maze_cmd_cb",
        "",
    )
